* ------------------------------------------------------------------------------
* stochastic energy systems model
*
* Volker Krey and Keywan Riahi
*
* International Institute for Applied Systems Analysis (IIASA), Schlossplatz 1,
* 2361 Laxenburg, AUSTRIA
* ------------------------------------------------------------------------------

* ------------------------------------------------------------------------------
* Options:
*
* %title% - title of model scenario or run (mandatory)
*
* %type% - type of optimization: 'deterministic' (default)
*                                'stochastic_linear_combination' - can be combined with all %formulation% settings (see below)
*                                'stochastic_cost_constraint' - can be combined with all %formulation% settings (see below)
*                                'stochastic_risk_constraint' - can be combined with all %formulation% settings (see below)
*                                'stochastic_linear_comb_cvar' - can be combined with all %formulation% settings (see below)
*                                'stochastic_cvar_constraint'
*                                'stochastic_cost_constraint_cvar'
*
* %formulation% - mathematical formulation: 'linear' (default)
*                                           'quadratic'
*                                           'linear_quadratic'
*
* %stochastic% - stochastic input file (name)
*
* %id% - additional identifier for model report files to distinguish scenarios/sensitivities
*
* ------------------------------------------------------------------------------
* EXAMPLES:
*
* From the command line a deterministic model run, called "global_test" is started with the following command:
* gams model_stochastic.gms --title=global_test --type=deterministic
*
* A stochastic cost-constrained model run with a linear risk measure, also called "global_test",
* with the sampled stochastic parameters in a file "stochastic_32_parameters_lognormal_lhs_N-1000_S-0.gms"
* is started with the command:
* gams model_stochastic.gms --title=global_test --type=stochastic_cost_constraint --formulation=linear --stochastic=stochastic_32_parameters_lognormal_lhs_N-1000_S-0
*
* If started from the GAMS IDE, just the options need to be entered into the command line
* ------------------------------------------------------------------------------

* ------------------------------------------------------------------------------
* check supplied options and defaults
* ------------------------------------------------------------------------------

* title is mandatory
$IF NOT set title $ERROR No title for model run supplied!

* "deterministic" is default optimization strategy type
$IF NOT set type $SETGLOBAL type deterministic

* "linear" is default formulation
$IF NOT set formulation $SETGLOBAL formulation FALSE

* "start" is default mode
$IF NOT set mode $SETGLOBAL mode start

* "false" is default user-file flag
$IF NOT set user $SETGLOBAL user FALSE

* "global" is default interest type
$IF NOT set interest $SETGLOBAL interest global

* ------------------------------------------------------------------------------
* model title
* ------------------------------------------------------------------------------

$IF NOT set id $TITLE "%title%_%type%_%formulation%"
$IF set id $TITLE "%title%_%id%_%type%_%formulation%"
$IF NOT set stochastic $TITLE "%title%_%type%"

* all model scenario specific files are stored in the following folder
$SETGLOBAL path "."

* ------------------------------------------------------------------------------
* GAMS options
* ------------------------------------------------------------------------------

* allow end-of-line and inline comments
$EOLCOM #
$INLINECOM /* */

* allow redefinition of sets and parameters
$ONMULTI

* allow empty sets
$ONEMPTY

* interpret 0 (zero) as small number (eps)
$ONEPS

* general GAMS solver options
$LABEL option_section

* suppress listing of model code and equations (avoid potentially huge .lst file due to stochastic model part)
$OFFLISTING
OPTION LIMROW = 0
OPTION LIMCOL = 0
* suppress listing of model solution (avoid potentially huge .lst file due to stochastic result part)
OPTION SOLPRINT = OFF

* solvers for linear and quadratic model versions
OPTION LP = CPLEX
OPTION QCP = CPLEX

* show control options
$SHOW

* ------------------------------------------------------------------------------
* model definition
* ------------------------------------------------------------------------------

* basic sets of model
$INCLUDE "set_definition.gms"

* model parameters
$INCLUDE "parameter_definition.gms"

* parameters for stochastic model parameters
$INCLUDE "stochastic_parameter.gms"

* stochastic input data
$IF set stochastic $INCLUDE "%stochastic%.gms"

* ------------------------------------------------------------------------------
* additional calculations that need to be done prior to defining equations
* ------------------------------------------------------------------------------

* ------------------------------------------------------------------------------
* aliases (needed e.g. for double summation)
* ------------------------------------------------------------------------------

ALIAS    (year, year_alias),
         (period, period_alias),
         (region, region_alias),
         (model_region, model_region_alias)
;

SCALAR period_counter    'period counter for loop' ;

SCALAR period_length_counter    'period length counter for loop' ;

* ------------------------------------------------------------------------------
* definition of period order
* ------------------------------------------------------------------------------

PARAMETER order(year)     'order for members of set year' ;
order(year)  = ORD(year) ;

* definition of first period
SET first_period(period) ;
first_period(period) = yes $ (ORD(period) eq 1) ;

* definition of last period
SET last_period(period) ;
last_period(period) = yes $ (ORD(period) eq CARD(period)) ;

* additional stochastic period sets
SET period_certain(period)  'periods with known parameters' ;
SET period_uncertain(period)  'periods with unknown parameters' ;

* definition of next period
SET next(period) ;

* definition of previous period
SET previous(period) ;

* future periods
SET future(year, year) ;

* past periods
SET past(year, year) ;

* past and present periods
SET past_presence(year, year) ;

* number of first period
period_counter = SMIN(period, order(period)) ;

* definition of next period
next(period) = yes $ (order(period) = period_counter + 1) ;

* definition of previous period
previous(period) = yes $ (order(period) = period_counter - 1) ;

* future periods
future(year, year_alias) = yes $ (order(year_alias) > order(year)) ;

* past periods
past(year, year_alias) = yes $ (order(year_alias) < order(year)) ;

* past and present periods
past_presence(year, year_alias) = yes $ (order(year_alias) <= order(year)) ;

* ------------------------------------------------------------------------------
* dynamic subsets for availability of technologies in regions and periods
* ------------------------------------------------------------------------------

SET technology_regional(period, model_region, technology, year)      'regional technologies'  ;
technology_regional(period, model_region, technology, year) = yes $
* technologies in past years available if they have been built, i.e. a capacity bound exists
    (past(period, year) and
     SUM(year_alias $ (order(year_alias) <= order(year)),
          technology_regional_start(model_region, technology, year_alias)) and
     (r_bound_capacity(period, model_region, technology, year, 'LO') > 0 or
      r_bound_capacity(period, model_region, technology, year, 'FX') > 0 or
      r_bound_capacity(period, model_region, technology, year, 'UP') > 0))
     or
* or in the present period if the technology is already commercially available
    (order(year) = order(period) and
     SUM(year_alias $ (order(year_alias) <= order(year)),
          technology_regional_start(model_region, technology, year_alias)))
;

* ------------------------------------------------------------------------------
* calculated input parameters
* ------------------------------------------------------------------------------

* regional economic data

PARAMETER cost_annual(model_region, technology, year)    'annualized investment costs' ;

cost_annual(model_region, technology, year) $
         (SUM(year_alias $ (order(year_alias) <= order(year)),
              technology_regional_start(model_region, technology, year_alias)) AND
          (NOT no_capacity(technology)) AND
          economic_lifetime(model_region, technology, year) > 0)
* sum is only a dummy, since each technology belongs exactly to one sector
         = SUM(sector $
                 (sectoral_technology(sector, technology) AND
                 regional_sector(model_region, sector)),
                 (  inv_cost(model_region, technology, year)                /* investment costs */
                  + (dec_cost(model_region, technology, year)               /* discounted decommissioning costs at end of TECHNICAL lifetime */
                    * (1 + interest(model_region, sector))**(-technical_lifetime(model_region, technology, year))))
               * ((1 + interest(model_region, sector))**(economic_lifetime(model_region, technology, year) - annuity_type)
                 * interest(model_region, sector))
               / ((1 + interest(model_region, sector))**(economic_lifetime(model_region, technology, year)) - 1))
;

* ------------------------------------------------------------------------------
* discount factors
* ------------------------------------------------------------------------------

* inter-periodical discount factor
PARAMETER discount_factor(period, model_region, sector)    'discount factor for inter-periodical discounting' ;

* as an approximation the discount rate is assumed to be the interest rate
discount_factor(period, model_region, sector) = (1 + interest(model_region, sector))**((order('2000') - order(period))*period_length);

* definition of set that has number of elements corresponding to number of years per period, i.e. period_length

* inner-periodical discount factor
PARAMETER discount_period(model_region, sector)            'discount factor for intra-periodical discounting' ;

discount_period(model_region, sector) = 0 ;

FOR (period_length_counter = 0 TO (period_length - 1) BY 1,
* discounting base year is the report year, i.e. either the first or last year of a period - therefore the exponent can be negative or positive, depending on the scalar "report_year"!
    discount_period(model_region, sector) = discount_period(model_region, sector) + (1 + interest(model_region, sector))**(report_year * period_length_counter) ;
) ;

$IF NOT set stochastic $GOTO parameter_derived_end

* ------------------------------------------------------------------------------
* model specifications - Stochastic MESSAGE Model
* ------------------------------------------------------------------------------

$LABEL stochastic_parameters

SET stochastic_scenario(stochastic_scenario_all)            'number of stochastic scenarios' ;

stochastic_scenario(stochastic_scenario_all) = YES $ (ORD(stochastic_scenario_all) <= stochastic_realization) ;

* calculate annualized stochastic investment costs

PARAMETER cost_annual_stoch(stochastic_scenario_all, model_region, technology, year)    'annualized stochastic investment costs' ;

cost_annual_stoch(stochastic_scenario, model_region, technology, year) $
         (SUM(year_alias $ (order(year_alias) <= order(year)),
              technology_regional_start(model_region, technology, year_alias)) AND
          (NOT no_capacity(technology)) AND
          economic_lifetime(model_region, technology, year) > 0)
* sum is only a dummy, since each technology belongs exactly to one sector
         = SUM(sector $
                 (sectoral_technology(sector, technology) AND
                 regional_sector(model_region, sector)),
                 (
                  + inv_cost_stoch(stochastic_scenario, model_region, technology, year) $ inv_cost_stoch(stochastic_scenario, model_region, technology, year)               /* investment costs */
                  + (dec_cost_stoch(stochastic_scenario, model_region, technology, year)               /* discounted decommissioning costs at end of TECHNICAL lifetime */
                    * (1 + interest(model_region, sector))**(-technical_lifetime(model_region, technology, year))) $ dec_cost_stoch(stochastic_scenario, model_region, technology, year)
                 )
               * ((1 + interest(model_region, sector))**(economic_lifetime(model_region, technology, year) - annuity_type)
                 * interest(model_region, sector))
               / ((1 + interest(model_region, sector))**(economic_lifetime(model_region, technology, year)) - 1))
;

* rename stochastic costs
inv_cost_stoch(stochastic_scenario_all, model_region, technology, year) = stinv(stochastic_scenario_all, model_region, technology, year) ;
var_cost_stoch(stochastic_scenario_all, model_region, technology, year) = stvom(stochastic_scenario_all, model_region, technology, year) ;
tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period) = sttax(stochastic_scenario, model_region, emission, period) ;

$LABEL parameter_derived_end

* ------------------------------------------------------------------------------
* equation and model definition
* ------------------------------------------------------------------------------

VARIABLES

* capacity and activity variables

         CAP(period, model_region, technology, year)                            'regional technology capacities per construction period'
         CAP_TOTAL(period, model_region, technology)                            'total regional technology capacity'
         ACT(period, model_region, technology, year)                            'annual activities for regional technologies per construction period'
         ACT_TOTAL(period, model_region, technology)                            'total annual activities for regional technologies'

         STOCK(period, model_region, energycarrier, level)                       'remaining stock of resources at beginning of period'

         ENERGY_DEMAND(period, model_region, energycarrier, level)               'energy demand on a certain level'
         ENERGY_SUPPLY(period, model_region, energycarrier, level)               'energy supply on a certain level'

* balance variables

         TOTAL_ENERGY_CONSUMPTION(period, energycarrier)                         'total energy consumption'
         TOTAL_ENERGY_PRODUCTION(period, energycarrier)                          'total energy production'

         REGIONAL_ENERGY_CONSUMPTION(period, model_region, energycarrier)        'energy consumption by region'
         REGIONAL_ENERGY_PRODUCTION(period, model_region, energycarrier)         'energy production by region'

         SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier)      'energy consumption by region/sector'
         SECTORAL_ENERGY_PRODUCTION(period, model_region, sector, energycarrier)       'energy production by region/sector'

         TOTAL_EMISSION(period, emission)                                        'total emission output'
         REGIONAL_EMISSION(period, model_region, emission)                       'emission output by region'
         SECTORAL_EMISSION(period, model_region, sector, emission)               'emission output by region/sector'

         TOTAL_COST(period)                                                      'total system costs by periods'
         REGIONAL_COST(period, model_region)                                     'costs by region'
         SECTORAL_COST(period, model_region, sector)                             'costs by region/sector'
         SECTORAL_COST_MESSAGE(period, model_region, sector)                     'costs by region/sector (MESSAGE accounting)'

         DEMAND_BALANCE(period, model_region, demand)                            'demand balance'
         DEMAND_POS_CONTRIBUTION(period, model_region, demand)                   'positive demand contribution from technologies'
         DEMAND_NEG_CONTRIBUTION(period, model_region, demand)                   'negative demand contribution from technologies'

         TOTAL_COST_PERFECT_FORESIGHT      'summed total discounted system costs for all periods'
         TOTAL_COST_MESSAGE
         TOTAL_COST_PERIOD_MESSAGE(period)
         CUMULATIVE_EMISSION(emission)     'cumulative total emissions'
;

POSITIVE VARIABLE
         CAP,
         CAP_TOTAL,
         ACT,
         ACT_TOTAL,
         STOCK,
         ENERGY_DEMAND,
         ENERGY_SUPPLY,
         REMAIN
;

EQUATIONS

* balance equations

         EQ_COST_TOTAL(period)                                                      'objective funtion by period'
         EQ_COST_REGIONAL(period, model_region)                                     'objective function by period/region'
         EQ_COST_SECTORAL(period, model_region, sector)                             'objective function by period/region/sector'
         EQ_COST_SECTORAL_MESSAGE(period, model_region, sector)                     'objective function by period/region/sector (MESSAGE accounting)'

         EQ_ENERGY_CONSUMPTION_TOTAL(period, energycarrier)                         'energy consumption'
         EQ_ENERGY_PRODUCTION_TOTAL(period, energycarrier)                          'energy production'

         EQ_ENERGY_CONSUMPTION_REGIONAL(period, model_region, energycarrier)      'energy consumption by region'
         EQ_ENERGY_PRODUCTION_REGIONAL(period, model_region, energycarrier)       'energy production by region'

         EQ_ENERGY_CONSUMPTION_SECTORAL(period, model_region, sector, energycarrier)      'energy consumption by region/sector'
         EQ_ENERGY_PRODUCTION_SECTORAL(period, model_region, sector, energycarrier)       'energy production by region/sector'

         EQ_EMISSION_TOTAL(period, emission)                                        'total emission output'
         EQ_EMISSION_REGIONAL(period, model_region, emission)                       'emission output by region'
         EQ_EMISSION_SECTORAL(period, model_region, sector, emission)               'emission output by region/sector'

* technology specific equations

         EQ_MAX_AV(period, model_region, technology, year)                          'maximum availability'

         EQ_DEMAND_SATISFACTION(period, model_region, demand)                       'demand satisfaction by region/demand category'
         EQ_DEMAND_POS_CONTRIBUTION(period, model_region, demand)                   'positive demand contribution from technologies'
         EQ_DEMAND_NEG_CONTRIBUTION(period, model_region, demand)                   'negative demand contribution from technologies'
         EQ_DEMAND_BALANCE(period, model_region, demand)                            'demand balance (summation of positive and negative contributions and exogeneous demand'

         EQ_UPGRADE_CAP(period, model_region, upgrade)                           'regional upgrade technology capacity constraint'
         EQ_UPGRADE_ACT(period, model_region, upgrade)                           'regional upgrade technology activity constraint'

         EQ_ENERGY_SUPPLY(period, model_region, energycarrier, level)              'supply of energycarriers'
         EQ_ENERGY_DEMAND(period, model_region, energycarrier, level)              'demand for energycarriers'
         EQ_ENERGY_BALANCE(period, model_region, energycarrier, level)             'supply > demand for energycarriers'

         EQ_CAPACITY_TOTAL(period, model_region, technology)                 'cumulated regional capacities (summed over construction year)'
         EQ_ACTIVITY_TOTAL(period, model_region, technology)                 'cumulated regional activities (summed over construction year)'

         EQ_COST_SUM_PERFECT_FORESIGHT                                              'objective function - total system costs for all periods'
         EQ_COST_TOTAL_MESSAGE
         EQ_COST_PERIOD_MESSAGE(period)
         EQ_CAP_COUPLING(period, model_region, technology, year)                    'capacity coupling for regional technologies'
         EQ_STOCK(period, model_region, energycarrier, level)                       'accounting of resource stock'
         EQ_EMISSION_CUMULATIVE(emission)                                           'accounting of cumulative emissions'
;

* ------------------------------------------------------------------------------
* costs
* ------------------------------------------------------------------------------

* all costs in the following equations are annual costs, summation and discounting is performed in EQ_COST_SUM_PERFECT_FORESIGHT

* total system costs for periods
EQ_COST_TOTAL(period)
..
      TOTAL_COST(period)
    - SUM(model_region,
            REGIONAL_COST(period, model_region))
      =E= 0
;

* regional costs for periods
EQ_COST_REGIONAL(period, model_region)
..
      REGIONAL_COST(period, model_region)
    - SUM(sector $ (regional_sector(model_region, sector)),
            SECTORAL_COST(period, model_region, sector))
      =E= 0
;

* sectoral costs for periods (without intra-periodical discounting, i.e. annual costs)
EQ_COST_SECTORAL(period, model_region, sector) $
regional_sector(model_region, sector)
..
      SECTORAL_COST(period, model_region, sector)
    - (
* investment- and fixcosts of regional technologies
      SUM((technology, year) $
          (sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           (NOT no_capacity(technology)) AND
           past_presence(period, year)),
            CAP(period, model_region, technology, year)
* annuity of investment costs of construction year (!) within economical lifetime (!) (SEE $ condition)
          * (cost_annual(model_region, technology, year) $ (economic_lifetime(model_region, technology, year) > (order(period) - order(year))*period_length)
* fixed O&M costs of construction year (!)
             + fix_cost(model_region, technology, year)))
* variable costs of regional technologies
    + SUM((technology, year) $
          (sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            ACT(period, model_region, technology, year)
          * var_cost(model_region, technology, year))
* taxes on energycarrier consumption (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM(energycarrier,
          (
              tax_energy_total(energycarrier, period)
            + tax_energy_regional(model_region, energycarrier, period)
            + tax_energy_sectoral(model_region, sector, energycarrier, period)
          )
          * SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier))
* taxes on emissions (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM(emission,
          (
               tax_emission_total(emission, period)
             + tax_emission_regional(model_region, emission, period)
             + tax_emission_sectoral(model_region, sector, emission, period)
          )
          * SECTORAL_EMISSION(period, model_region, sector, emission))
      )
    =E= 0
;

*$ONTEXT
* sectoral costs for periods (according to MESSAGE - with intra-periodical discounting))
EQ_COST_SECTORAL_MESSAGE(period, model_region, sector) $
regional_sector(model_region, sector)
..
      SECTORAL_COST_MESSAGE(period, model_region, sector)
    - (
* investment- and fixcosts of regional technologies
      SUM((technology, year) $
          (sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           (NOT no_capacity(technology)) AND
           past_presence(period, year)),
            (
            CAP(period, model_region, technology, year)
* total investment costs
            * cost_annual(model_region, technology, year)
* construction time of five years: investment costs discounted to middle of construction period
            * (1 + interest(model_region, sector))**(2.5)
* investment costs discounted to middle of period (report_year dependent)
            * (1 + interest(model_region, sector))**(period_length/2 * report_year)
* factor corrects for cut-off effect of finite time horizon (card(year) is equivalent to order(last_period) possible!)
            * ((1 + interest(model_region, sector))**min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - 1)
            / ((1 + interest(model_region, sector))**(min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - annuity_type) * interest(model_region, sector))
            ) $ sameas(period, year)
* fixed O&M costs of construction year (!)
             + CAP(period, model_region, technology, year)
             * fix_cost(model_region, technology, year)
             * discount_period(model_region, sector))
* variable costs of regional technologies
    + SUM((technology, year) $
          (sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            ACT(period, model_region, technology, year)
          * var_cost(model_region, technology, year)
* MESSAGE discounting for "vom": period_length * disounting to middle of period
*          * period_length * (1 + interest(model_region, sector))**(-period_length/2)
* Standard discounting: each year separately discounted
          * discount_period(model_region, sector)
         )
* taxes on energycarrier consumption (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM(energycarrier,
          (
              tax_energy_total(energycarrier, period)
            + tax_energy_regional(model_region, energycarrier, period)
            + tax_energy_sectoral(model_region, sector, energycarrier, period)
          )
          * SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier)
* MESSAGE discounting for "vom": period_length * disounting to middle of period
*          * period_length * (1 + interest(model_region, sector))**(-period_length/2)
* Standard discounting: each year separately discounted
          * discount_period(model_region, sector)
         )
* taxes on emissions (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM(emission,
          (
               tax_emission_total(emission, period)
             + tax_emission_regional(model_region, emission, period)
             + tax_emission_sectoral(model_region, sector, emission, period)
          )
          * SECTORAL_EMISSION(period, model_region, sector, emission)
* MESSAGE discounting for "vom": period_length * disounting to middle of period
*          * period_length * (1 + interest(model_region, sector))**(-period_length/2)
* Standard discounting: each year separately discounted
          * discount_period(model_region, sector)
         )
      )
    =E= 0
;

* ------------------------------------------------------------------------------
* energy consumption
* ------------------------------------------------------------------------------

* total energy consumption
EQ_ENERGY_CONSUMPTION_TOTAL(period, energycarrier)
..
      TOTAL_ENERGY_CONSUMPTION(period, energycarrier)
    - SUM(model_region,
            REGIONAL_ENERGY_CONSUMPTION(period, model_region, energycarrier))
      =E= 0
;

* regional energy consumption
EQ_ENERGY_CONSUMPTION_REGIONAL(period, model_region, energycarrier)
..
      REGIONAL_ENERGY_CONSUMPTION(period, model_region, energycarrier)
    - SUM(sector $ regional_sector(model_region, sector),
            SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier))
      =E= 0
;

* sectoral energy consumption
EQ_ENERGY_CONSUMPTION_SECTORAL(period, model_region, sector, energycarrier) $
regional_sector(model_region, sector)
..
      SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier)
* energy consumption of regional technologies
    - SUM((technology, year, level) $
          (sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year) AND energycarrier_level(energycarrier, level)),
            input(model_region, technology, energycarrier, level, year)
          * ACT(period, model_region, technology, year))
      =E= 0
;

* ------------------------------------------------------------------------------
* energy production
* ------------------------------------------------------------------------------

* total energy production
EQ_ENERGY_PRODUCTION_TOTAL(period, energycarrier)
..
      TOTAL_ENERGY_PRODUCTION(period, energycarrier)
    - SUM(model_region,
            REGIONAL_ENERGY_PRODUCTION(period, model_region, energycarrier))
      =E= 0
;

* regional energy production
EQ_ENERGY_PRODUCTION_REGIONAL(period, model_region, energycarrier)
..
      REGIONAL_ENERGY_PRODUCTION(period, model_region, energycarrier)
    - SUM(sector $ regional_sector(model_region, sector),
            SECTORAL_ENERGY_PRODUCTION(period, model_region, sector, energycarrier))
      =E= 0
;

* sectoral energy production
EQ_ENERGY_PRODUCTION_SECTORAL(period, model_region, sector, energycarrier) $
regional_sector(model_region, sector)
..
      SECTORAL_ENERGY_PRODUCTION(period, model_region, sector, energycarrier)
* energy consumption of regional technologies
    - SUM((technology, year, level) $
          (sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year) AND energycarrier_level(energycarrier, level)),
            output(model_region, technology, energycarrier, level, year)
          * ACT(period, model_region, technology, year))
      =E= 0
;

* ------------------------------------------------------------------------------
* emissions
* ------------------------------------------------------------------------------

* total emissions
EQ_EMISSION_TOTAL(period, emission)
..
      TOTAL_EMISSION(period, emission)
    - SUM(model_region,
            REGIONAL_EMISSION(period, model_region, emission))
      =E= 0
;

* emissions of regions
EQ_EMISSION_REGIONAL(period, model_region, emission)
..
      REGIONAL_EMISSION(period, model_region, emission)
* emissions of regional technologies
    - SUM(sector $ regional_sector(model_region, sector),
            SECTORAL_EMISSION(period, model_region, sector, emission))
      =E= 0
;

* emissions of sectors
EQ_EMISSION_SECTORAL(period, model_region, sector, emission) $
regional_sector(model_region, sector)
..
      SECTORAL_EMISSION(period, model_region, sector, emission)
* emissions of regional technologies
    - SUM((technology, year) $
          (sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            ACT(period, model_region, technology, year)
          * emission_output(model_region, technology, emission, year))
      =E= 0
;

* ------------------------------------------------------------------------------
* technology specific equations
* ------------------------------------------------------------------------------

* availability constraint for regional technologies
EQ_MAX_AV(period, model_region, technology, year) $
* equation only to be created for existing model_region - technology combinations
(technology_regional(period, model_region, technology, year) AND
 (NOT no_capacity(technology)) AND past_presence(period, year))
..
      CAP(period, model_region, technology, year)
    * availability(model_region, technology, year)
    - ACT(period, model_region, technology, year)
      =G= 0
;

* ------------------------------------------------------------------------------
* equations for upgrade technologies
* ------------------------------------------------------------------------------

* upgrade technology capacity constraint for regional technologies
EQ_UPGRADE_CAP(period, model_region, upgrade) $
* equation only to be created for existing model_region - technology combinations
technology_regional(period, model_region, upgrade, period)
..
* cumulated installed capacity of corresponding parent-technology
      SUM(technology $ upgrade_correspondence(technology, upgrade),
          CAP_TOTAL(period, model_region, technology))
* cumulated installed capacity of upgrade technology
        - CAP_TOTAL(period, model_region, upgrade)
      =G= 0
;

* upgrade technology activity constraint for regional technologies
EQ_UPGRADE_ACT(period, model_region, upgrade) $
* equation only to be created for existing model_region - technology combinations
technology_regional(period, model_region, upgrade, period)

..
* cumulated activity of corresponding parent-technology
      SUM(technology $ upgrade_correspondence(technology, upgrade),
          ACT_TOTAL(period, model_region, technology))
* cumulated activity of upgrade technology
    - ACT_TOTAL(period, model_region, upgrade)
      =G= 0
;

* ------------------------------------------------------------------------------
* energy-service demand specific equations
* ------------------------------------------------------------------------------

* demand balance has to be positive (i.e. demands have to be satisfied)
EQ_DEMAND_SATISFACTION(period, model_region, demand)
..
      DEMAND_BALANCE(period, model_region, demand)
      =G= 0
;

* demand satisfied by demand technologies (positive demand contribution from technologies)
EQ_DEMAND_POS_CONTRIBUTION(period, model_region, demand)
..
      DEMAND_POS_CONTRIBUTION(period, model_region, demand)
* regional technologies
    - SUM((demand_technology, year) $
          (technology_regional(period, model_region, demand_technology, year) AND
           demand_correspondence(model_region, demand_technology, demand) AND
           past_presence(period, year)),
            demand_output(model_region, demand_technology, demand, year)
          * ACT(period, model_region, demand_technology, year))
      =E= 0
;

* demand created by demand technologies (negative demand contribution from technologies)
EQ_DEMAND_NEG_CONTRIBUTION(period, model_region, demand)
..
      DEMAND_NEG_CONTRIBUTION(period, model_region, demand)
* regional technologies
    - SUM((demand_technology, year) $
          (technology_regional(period, model_region, demand_technology, year) AND
           demand_correspondence(model_region, demand_technology, demand) AND
           past_presence(period, year)),
            demand_input(model_region, demand_technology, demand, year)
          * ACT(period, model_region, demand_technology, year))
      =E= 0
;

* summation of different demand contributions (direct and indirect - new since v4)
EQ_DEMAND_BALANCE(period, model_region, demand)
..
      DEMAND_BALANCE(period, model_region, demand)
* direct contributions to demand
* demand satisfied by demand technologies
    - DEMAND_POS_CONTRIBUTION(period, model_region, demand)
* demand created by demand technologies
    + DEMAND_NEG_CONTRIBUTION(period, model_region, demand)
* exogenous demand
    + demand_regional(model_region, demand, period)
* indirect contributions from subdemands
*    - SUM(subdemand $ demand_structure(demand, subdemand),
*          DEMAND_BALANCE(period, model_region, subdemand))
      =E= 0
;

* ------------------------------------------------------------------------------
* energy markets: supply and demand
* ------------------------------------------------------------------------------

* supply of energycarriers due to technologies and import
EQ_ENERGY_SUPPLY(period, model_region, energycarrier, level) $
energycarrier_level(energycarrier, level)
..
      ENERGY_SUPPLY(period, model_region, energycarrier, level)
* output of regional technologies
    - SUM((technology, year) $
          (technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            output(model_region, technology, energycarrier, level, year)
          * ACT(period, model_region, technology, year))
* resource extraction (with input! - balance is taken care of in equation EQ_STOCK)
    - (SUM((technology, year) $ technology_regional(period, model_region, technology, year),
            input(model_region, technology, energycarrier, level, year)
          * ACT(period, model_region, technology, year))) $ volume(model_region, energycarrier, level)
      =E= 0
;

* demand for non-grid-bound energycarriers (including exports)
* exports to world market cannot be considered, since some i
EQ_ENERGY_DEMAND(period, model_region, energycarrier, level) $
energycarrier_level(energycarrier, level)
..
      ENERGY_DEMAND(period, model_region, energycarrier, level)
* input of technologies and regional networks
    - SUM((technology, year) $
          (technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            input(model_region, technology, energycarrier, level, year)
          * ACT(period, model_region, technology, year))
      =E= 0
;

* supply has to be larger than demand
EQ_ENERGY_BALANCE(period, model_region, energycarrier, level) $
energycarrier_level(energycarrier, level)
..
      ENERGY_SUPPLY(period, model_region, energycarrier, level)
    - ENERGY_DEMAND(period, model_region, energycarrier, level)
      =G= 0
;

* ------------------------------------------------------------------------------
* restriction on share of intermittent sources
* ------------------------------------------------------------------------------
EQUATION
      EQ_BASELOAD(period, model_region)         'restriction on share of baseload electricity generating sources'
      EQ_INTERMITTENT(period, model_region)     'restriction on share of intermittent electricity generating sources'
;

EQ_BASELOAD(period, model_region)..
      0.6 * ACT_TOTAL(period, model_region, 'enduse_electric_std')
* end-use demand (VK, 25/06/2208: changed, because reduces incentive to invest into electricity savings)
*      0.6 * (ACT_TOTAL(period, model_region, 'enduse_electric_std') + ACT_TOTAL(period, model_region, 'enduse_electric_save1') + ACT_TOTAL(period, model_region, 'enduse_electric_save2'))
* conversion sector demand
    + 1.0 * SUM(technology $ ((NOT SAMEAS(technology, 'elec_t_d')) AND (sectoral_technology('conversion', technology) OR sectoral_technology('primary_energy', technology))),
                ACT_TOTAL(period, model_region, technology) * input(model_region, technology, 'electricity', 'secondary', period))
* baseload technologies
    - ACT_TOTAL(period, model_region, 'nuc_hc')
      =G= 0
;

EQ_INTERMITTENT(period, model_region)..
      0.25 * ACT_TOTAL(period, model_region, 'enduse_electric_std')
* end-use demand (VK, 25/06/2208: changed, because reduces incentive to invest into electricity savings)
*      0.25 * (ACT_TOTAL(period, model_region, 'enduse_electric_std') + ACT_TOTAL(period, model_region, 'enduse_electric_save1') + ACT_TOTAL(period, model_region, 'enduse_electric_save2'))
* conversion sector demand
    + 0.75 * SUM(technology $ ((NOT SAMEAS(technology, 'elec_t_d')) AND (sectoral_technology('conversion', technology) OR sectoral_technology('primary_energy', technology))),
                ACT_TOTAL(period, model_region, technology) * input(model_region, technology, 'electricity', 'secondary', period))
* backup
    + ACT_TOTAL(period, model_region, 'backup_elc')
* intermittent sources (excl. hydro because of potential storage option)
    - ACT_TOTAL(period, model_region, 'wind_ppl')
    - ACT_TOTAL(period, model_region, 'wind_ppl2')
    - ACT_TOTAL(period, model_region, 'solar_pv_ppl')
      =G= 0
;

* ------------------------------------------------------------------------------
* total capacity and activity per technology (summation over construction periods)
* ------------------------------------------------------------------------------

* regional total capacity (summation over construction periods)
EQ_CAPACITY_TOTAL(period, model_region, technology) $
(technology_regional(period, model_region, technology, period) AND
 (NOT no_capacity(technology)))
..
      CAP_TOTAL(period, model_region, technology)
    - SUM (year $
           (technology_regional(period, model_region, technology, year) AND
            past_presence(period, year)),
              CAP(period, model_region, technology, year))
      =E= 0
;

* regional total activity (summation over construction periods)
EQ_ACTIVITY_TOTAL(period, model_region, technology) $
technology_regional(period, model_region, technology, period)
..
      ACT_TOTAL(period, model_region, technology)
    - SUM (year $
           (technology_regional(period, model_region, technology, year) AND
            past_presence(period, year)),
              ACT(period, model_region, technology, year))
      =E= 0
;

* ------------------------------------------------------------------------------
* model specifications - Perfect-Foresight Model
* ------------------------------------------------------------------------------
$LABEL perfect_foresight_equation

* summed total system costs for perfect foresight optimization, i.e. total discounted costs for all periods
EQ_COST_SUM_PERFECT_FORESIGHT
..
      TOTAL_COST_PERFECT_FORESIGHT
    - SUM((period, model_region, sector) $ regional_sector(model_region, sector),
            SECTORAL_COST(period, model_region, sector)
* discounting for years in single period
          * discount_period(model_region, sector)
* discounting for different periods
          * discount_factor(period, model_region, sector))
* carbon price on cumulative emission for last period
    - SUM((emission, last_period), CUMULATIVE_EMISSION(emission)
          * tax_emission_cumulative(emission)
          * discount_factor(last_period, 'world', 'primary_energy'))
      =E= 0
;

EQ_COST_TOTAL_MESSAGE
..
      TOTAL_COST_MESSAGE
    - SUM((period, model_region, sector) $ regional_sector(model_region, sector),
          SECTORAL_COST_MESSAGE(period, model_region, sector) * discount_factor(period, model_region, sector))
* carbon price on cumulative emission for last period
    - SUM((emission, last_period),
            CUMULATIVE_EMISSION(emission)
          * tax_emission_cumulative(emission)
* MESSAGE discounting for "vom": disounting to middle of period
*          * (1 + interest('world', 'primary_energy'))**(-period_length/2)
          * discount_factor(last_period, 'world', 'primary_energy'))
      =E= 0
;

EQ_COST_PERIOD_MESSAGE(period)
..
      TOTAL_COST_PERIOD_MESSAGE(period)
    - SUM((model_region, sector) $ regional_sector(model_region, sector),
          SECTORAL_COST_MESSAGE(period, model_region, sector) * discount_factor(period, model_region, sector))
* carbon price on cumulative emission for last period
    - SUM((emission, last_period) $ SAMEAS(period, last_period),
            CUMULATIVE_EMISSION(emission)
          * tax_emission_cumulative(emission)
* MESSAGE discounting for "vom": disounting to middle of period
*          * (1 + interest('world', 'primary_energy'))**(-period_length/2)
*          * discount_factor(last_period, 'world', 'primary_energy')
      )
      =E= 0
;

* stock of resources
EQ_STOCK(period, model_region, energycarrier, level) $
(volume(model_region, energycarrier, level))
* AND (NOT last_period(period)))
..
      STOCK(period + 1, model_region, energycarrier, level)
    + REMAIN(model_region, energycarrier, level) $ last_period(period)
    - STOCK(period, model_region, energycarrier, level)
    + SUM((technology, year) $ technology_regional(period, model_region, technology, year),
            input(model_region, technology, energycarrier, level, year)
          * ACT(period, model_region, technology, year)) * period_length
      =E= 0
;

* cumulative emissions
EQ_EMISSION_CUMULATIVE(emission)
..
      CUMULATIVE_EMISSION(emission)
    - SUM(period, TOTAL_EMISSION(period, emission) * period_length)
      =E= 0
;


EQ_CAP_COUPLING(period, model_region, technology, year) $
(technology_regional(period, model_region, technology, year) AND
 technology_regional(period + 1, model_region, technology, year) AND
 (NOT no_capacity(technology)) AND
 (order(period) >= order(year)) AND
 (NOT last_period(period)))
..
* modification for lifetime that is not a multiple of period_length
         (1 * CAP(period, model_region, technology, year)) $ (technical_lifetime(model_region, technology, year) - (order(period + 1) + 1 - order(year))*period_length >= 0)
       + ((technical_lifetime(model_region, technology, year) - (order(period + 1) - order(year))*period_length)/period_length
            * CAP(period, model_region, technology, year)) $ ((technical_lifetime(model_region, technology, year) - (order(period + 1) + 1 - order(year))*period_length < 0) AND
                                                              (technical_lifetime(model_region, technology, year) - (order(period + 1) - order(year))*period_length > 0))
       - CAP(period + 1, model_region, technology, year)
       =E= 0
;

MODEL DETERMINISTIC /ALL/ ;

$IFI "%type%" == "deterministic" $GOTO end_equation

$IFI "%type%" == "stochastic_linear_combination" $GOTO stochastic_equation
$IFI "%type%" == "stochastic_cost_constraint" $GOTO stochastic_equation
$IFI "%type%" == "stochastic_risk_constraint" $GOTO stochastic_equation
$IFI "%type%" == "stochastic_linear_comb_cvar" $GOTO stochastic_equation
$IFI "%type%" == "stochastic_cvar_constraint" $GOTO stochastic_equation
$IFI "%type%" == "stochastic_cost_constraint_cvar" $GOTO stochastic_equation
$IFI "%type%" == "stochastic_fuzzy" $GOTO stochastic_equation

* ------------------------------------------------------------------------------
* Stochastic Model Specifications
* ------------------------------------------------------------------------------
$LABEL stochastic_equation

* additional variables
VARIABLE
    RISK_FUNCTION(period, stochastic_scenario_all)                 'sample risk function'
    OPPORTUNITY_FUNCTION(period, stochastic_scenario_all)          'sample opportunity function'
    RISK_FUNCTION_TOTAL                                            'total risk function'
    OPPORTUNITY_FUNCTION_TOTAL                                     'total opportunity function'
;

POSITIVE VARIABLE
    RISK_FUNCTION
;

* additional equations
EQUATION
    EQ_RISK_FUNCTION(period, stochastic_scenario_all)
    EQ_OPPORTUNITY_FUNCTION(period, stochastic_scenario_all)
    EQ_RISK_FUNCTION_TOTAL
    EQ_OPPORTUNITY_FUNCTION_TOTAL
;

$IFI "%formulation%" == "linear" $GOTO risk_function_linear
$IFI "%formulation%" == "quadratic" $GOTO risk_function_linear
$IFI "%formulation%" == "linear_quadratic" $GOTO risk_function_linear_quadratic

$LABEL risk_function_linear

EQ_RISK_FUNCTION(period, stochastic_scenario) $
(NOT first_period(period))
..
      RISK_FUNCTION(period, stochastic_scenario)
    - (
      SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           (NOT no_capacity(technology)) AND
           past_presence(period, year)),
            (
            CAP(period, model_region, technology, year)
* difference of stochastic and expexted value total investment costs
            * (cost_annual_stoch(stochastic_scenario, model_region, technology, year) - cost_annual(model_region, technology, year)) $ cost_annual_stoch(stochastic_scenario, model_region, technology, year)
* building time of five years: investment costs discounted to middle of construction period
            * (1 + interest(model_region, sector))**(2.5)
* investment costs discounted to middle of period (report_year dependent)
            * (1 + interest(model_region, sector))**(period_length/2 * report_year)
* factor corrects for cut-off effect of finite time horizon (card(year) is equivalent to order(last_period) possible!)
            * ((1 + interest(model_region, sector))**min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - 1)
            / ((1 + interest(model_region, sector))**(min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - annuity_type) * interest(model_region, sector))
            ) $ sameas(period, year)
* fixed O&M costs of construction year (!)
           + CAP(period, model_region, technology, year)
             * (fix_cost_stoch(stochastic_scenario, model_region, technology, year) - fix_cost(model_region, technology, year)) $ fix_cost_stoch(stochastic_scenario, model_region, technology, year)
             * discount_period(model_region, sector)
         )
* variable costs of regional technologies
    + SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            ACT(period, model_region, technology, year)
          * (var_cost_stoch(stochastic_scenario, model_region, technology, year) - var_cost(model_region, technology, year)) $ var_cost_stoch(stochastic_scenario, model_region, technology, year)
* MESSAGE discounting for "vom": period_length * disounting to middle of period
*          * period_length * (1 + interest(model_region, sector))**(-period_length/2)
* Standard discounting: each year separately discounted
          * discount_period(model_region, sector)
         )
* taxes on energycarrier consumption (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, energycarrier) $ regional_sector(model_region, sector),
          (
              (tax_energy_total_stoch(stochastic_scenario, energycarrier, period) - tax_energy_total(energycarrier, period)) $ tax_energy_total_stoch(stochastic_scenario, energycarrier, period)
            + (tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period) - tax_energy_regional(model_region, energycarrier, period)) $ tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period)
            + (tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period) - tax_energy_sectoral(model_region, sector, energycarrier, period)) $ tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period)
          )
          * SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier)
* MESSAGE discounting for "vom": period_length * disounting to middle of period
*          * period_length * (1 + interest(model_region, sector))**(-period_length/2)
* Standard discounting: each year separately discounted
          * discount_period(model_region, sector)
         )
* taxes on emissions (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, emission) $ regional_sector(model_region, sector),
          (
               (tax_emission_total_stoch(stochastic_scenario, emission, period) - tax_emission_total(emission, period)) $ tax_emission_total_stoch(stochastic_scenario, emission, period)
             + (tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period) - tax_emission_regional(model_region, emission, period)) $ tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period)
             + (tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period) - tax_emission_sectoral(model_region, sector, emission, period)) $ tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period)
          )
          * SECTORAL_EMISSION(period, model_region, sector, emission)
* MESSAGE discounting for "vom": period_length * disounting to middle of period
*          * period_length * (1 + interest(model_region, sector))**(-period_length/2)
* Standard discounting: each year separately discounted
          * discount_period(model_region, sector)
         )
* carbon price on cumulative emission for last period !!!
    + SUM(emission,
          (
                (tax_emission_cumulative_stoch(stochastic_scenario, emission) - tax_emission_cumulative(emission)) $ tax_emission_cumulative_stoch(stochastic_scenario, emission)
                * CUMULATIVE_EMISSION(emission)
* MESSAGE discounting for "vom": disounting to middle of period
*                * (1 + interest('world', 'primary_energy'))**(-period_length/2)
          ) $ last_period(period)
         )
      )
      =G= 0
;

EQ_OPPORTUNITY_FUNCTION(period, stochastic_scenario) $
(NOT first_period(period))
..
      OPPORTUNITY_FUNCTION(period, stochastic_scenario)
    - (
      SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           (NOT no_capacity(technology)) AND
           past_presence(period, year)),
            (
            CAP(period, model_region, technology, year)
* difference of stochastic and expexted value total investment costs
            * (cost_annual_stoch(stochastic_scenario, model_region, technology, year) - cost_annual(model_region, technology, year)) $ cost_annual_stoch(stochastic_scenario, model_region, technology, year)
* building time of five years: investment costs discounted to middle of construction period
            * (1 + interest(model_region, sector))**(2.5)
* investment costs discounted to middle of period (report_year dependent)
            * (1 + interest(model_region, sector))**(period_length/2 * report_year)
* factor corrects for cut-off effect of finite time horizon (card(year) is equivalent to order(last_period) possible!)
            * ((1 + interest(model_region, sector))**min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - 1)
            / ((1 + interest(model_region, sector))**(min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - annuity_type) * interest(model_region, sector))
            ) $ sameas(period, year)
* fixed O&M costs of construction year (!)
           + CAP(period, model_region, technology, year)
             * (fix_cost_stoch(stochastic_scenario, model_region, technology, year) - fix_cost(model_region, technology, year)) $ fix_cost_stoch(stochastic_scenario, model_region, technology, year)
             * discount_period(model_region, sector)
         )
* variable costs of regional technologies
    + SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            ACT(period, model_region, technology, year)
          * (var_cost_stoch(stochastic_scenario, model_region, technology, year) - var_cost(model_region, technology, year)) $ var_cost_stoch(stochastic_scenario, model_region, technology, year)
          * discount_period(model_region, sector)
         )
* taxes on energycarrier consumption (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, energycarrier) $ regional_sector(model_region, sector),
          (
              (tax_energy_total_stoch(stochastic_scenario, energycarrier, period) - tax_energy_total(energycarrier, period)) $ tax_energy_total_stoch(stochastic_scenario, energycarrier, period)
            + (tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period) - tax_energy_regional(model_region, energycarrier, period)) $ tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period)
            + (tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period) - tax_energy_sectoral(model_region, sector, energycarrier, period)) $ tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period)
          )
          * SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier)
          * discount_period(model_region, sector)
         )
* taxes on emissions (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, emission) $ regional_sector(model_region, sector),
          (
               (tax_emission_total_stoch(stochastic_scenario, emission, period) - tax_emission_total(emission, period)) $ tax_emission_total_stoch(stochastic_scenario, emission, period)
             + (tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period) - tax_emission_regional(model_region, emission, period)) $ tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period)
             + (tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period) - tax_emission_sectoral(model_region, sector, emission, period)) $ tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period)
          )
          * SECTORAL_EMISSION(period, model_region, sector, emission)
          * discount_period(model_region, sector)
         )
* carbon price on cumulative emission for last period !!!
    + SUM(emission,
          (
                (tax_emission_cumulative_stoch(stochastic_scenario, emission) - tax_emission_cumulative(emission)) $ tax_emission_cumulative_stoch(stochastic_scenario, emission)
                * CUMULATIVE_EMISSION(emission)
          ) $ last_period(period)
         )
      )
      =G= 0
;

$IFI "%formulation%" == "linear" $GOTO total_risk_function_linear
$IFI "%formulation%" == "quadratic" $GOTO total_risk_function_quadratic

$LABEL risk_function_linear_quadratic

EQ_RISK_FUNCTION(period, stochastic_scenario) $
(NOT first_period(period))
..
      RISK_FUNCTION(period, stochastic_scenario)
    - (
      SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           (NOT no_capacity(technology)) AND
           past_presence(period, year)),
            (
            CAP(period, model_region, technology, year)
* difference of stochastic and expexted value total investment costs
            * (cost_annual_stoch(stochastic_scenario, model_region, technology, year) - cost_annual(model_region, technology, year)) $ cost_annual_stoch(stochastic_scenario, model_region, technology, year)
* building time of five years: investment costs discounted to middle of construction period
            * (1 + interest(model_region, sector))**(2.5)
* investment costs discounted to middle of period (report_year dependent)
            * (1 + interest(model_region, sector))**(period_length/2 * report_year)
* factor corrects for cut-off effect of finite time horizon (card(year) is equivalent to order(last_period) possible!)
            * ((1 + interest(model_region, sector))**min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - 1)
            / ((1 + interest(model_region, sector))**(min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - annuity_type) * interest(model_region, sector))
            ) $ sameas(period, year)
* fixed O&M costs of construction year (!)
           + CAP(period, model_region, technology, year)
             * (fix_cost_stoch(stochastic_scenario, model_region, technology, year) - fix_cost(model_region, technology, year)) $ fix_cost_stoch(stochastic_scenario, model_region, technology, year)
             * discount_period(model_region, sector)
         )
* variable costs of regional technologies
    + SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            ACT(period, model_region, technology, year)
          * (var_cost_stoch(stochastic_scenario, model_region, technology, year) - var_cost(model_region, technology, year)) $ var_cost_stoch(stochastic_scenario, model_region, technology, year)
          * discount_period(model_region, sector)
         )
* taxes on energycarrier consumption (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, energycarrier) $ regional_sector(model_region, sector),
          (
              (tax_energy_total_stoch(stochastic_scenario, energycarrier, period) - tax_energy_total(energycarrier, period)) $ tax_energy_total_stoch(stochastic_scenario, energycarrier, period)
            + (tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period) - tax_energy_regional(model_region, energycarrier, period)) $ tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period)
            + (tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period) - tax_energy_sectoral(model_region, sector, energycarrier, period)) $ tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period)
          )
          * SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier)
          * discount_period(model_region, sector)
         )
* taxes on emissions (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, emission) $ regional_sector(model_region, sector),
          (
               (tax_emission_total_stoch(stochastic_scenario, emission, period) - tax_emission_total(emission, period)) $ tax_emission_total_stoch(stochastic_scenario, emission, period)
             + (tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period) - tax_emission_regional(model_region, emission, period)) $ tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period)
             + (tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period) - tax_emission_sectoral(model_region, sector, emission, period)) $ tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period)
          )
          * SECTORAL_EMISSION(period, model_region, sector, emission)
          * discount_period(model_region, sector)
         )
* carbon price on cumulative emission for last period !!!
    + SUM(emission,
          (
                (tax_emission_cumulative_stoch(stochastic_scenario, emission) - tax_emission_cumulative(emission)) $ tax_emission_cumulative_stoch(stochastic_scenario, emission)
                * CUMULATIVE_EMISSION(emission)
          ) $ last_period(period)
         )
      )
      =G= 0
;

EQ_OPPORTUNITY_FUNCTION(period, stochastic_scenario) $
(NOT first_period(period))
..
      OPPORTUNITY_FUNCTION(period, stochastic_scenario)
    - (
      SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           (NOT no_capacity(technology)) AND
           past_presence(period, year)),
            (
            CAP(period, model_region, technology, year)
* difference of stochastic and expexted value total investment costs
            * (cost_annual_stoch(stochastic_scenario, model_region, technology, year) - cost_annual(model_region, technology, year)) $ (cost_annual_stoch(stochastic_scenario, model_region, technology, year) AND (cost_annual_stoch(stochastic_scenario, model_region, technology, year) < cost_annual(model_region, technology, year)))
* building time of five years: investment costs discounted to middle of construction period
            * (1 + interest(model_region, sector))**(2.5)
* investment costs discounted to middle of period (report_year dependent)
            * (1 + interest(model_region, sector))**(period_length/2 * report_year)
* factor corrects for cut-off effect of finite time horizon (card(year) is equivalent to order(last_period) possible!)
            * ((1 + interest(model_region, sector))**min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - 1)
            / ((1 + interest(model_region, sector))**(min(economic_lifetime(model_region, technology, year), period_length * (card(year)-order(year) + 1)) - annuity_type) * interest(model_region, sector))
            ) $ sameas(period, year)
* fixed O&M costs of construction year (!)
           + CAP(period, model_region, technology, year)
             * (fix_cost_stoch(stochastic_scenario, model_region, technology, year) - fix_cost(model_region, technology, year)) $ (fix_cost_stoch(stochastic_scenario, model_region, technology, year) AND (fix_cost_stoch(stochastic_scenario, model_region, technology, year) < fix_cost(model_region, technology, year)))
             * discount_period(model_region, sector)
         )
* variable costs of regional technologies
    + SUM((model_region, sector, technology, year) $
          (regional_sector(model_region, sector) AND
           sectoral_technology(sector, technology) AND
           technology_regional(period, model_region, technology, year) AND
           past_presence(period, year)),
            ACT(period, model_region, technology, year)
          * (var_cost_stoch(stochastic_scenario, model_region, technology, year) - var_cost(model_region, technology, year)) $ (var_cost_stoch(stochastic_scenario, model_region, technology, year) AND (var_cost_stoch(stochastic_scenario, model_region, technology, year) < var_cost(model_region, technology, year)))
          * discount_period(model_region, sector)
         )
* taxes on energycarrier consumption (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, energycarrier) $ regional_sector(model_region, sector),
          (
              (tax_energy_total_stoch(stochastic_scenario, energycarrier, period) - tax_energy_total(energycarrier, period)) $ (tax_energy_total_stoch(stochastic_scenario, energycarrier, period) AND (tax_energy_total_stoch(stochastic_scenario, energycarrier, period) < tax_energy_total(energycarrier, period)))
            + (tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period) - tax_energy_regional(model_region, energycarrier, period)) $ (tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period) AND (tax_energy_regional_stoch(stochastic_scenario, model_region, energycarrier, period) < tax_energy_regional(model_region, energycarrier, period)))
            + (tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period) - tax_energy_sectoral(model_region, sector, energycarrier, period)) $ (tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period) AND (tax_energy_sectoral_stoch(stochastic_scenario, model_region, sector, energycarrier, period) < tax_energy_sectoral(model_region, sector, energycarrier, period)))
          )
          * SECTORAL_ENERGY_CONSUMPTION(period, model_region, sector, energycarrier)
          * discount_period(model_region, sector)
         )
* taxes on emissions (ATTENTION: implemented on different levels - avoid double counting!)
    + SUM((model_region, sector, emission) $ regional_sector(model_region, sector),
          (
               (tax_emission_total_stoch(stochastic_scenario, emission, period) - tax_emission_total(emission, period)) $ (tax_emission_total_stoch(stochastic_scenario, emission, period) AND (tax_emission_total_stoch(stochastic_scenario, emission, period) < tax_emission_total(emission, period)))
             + (tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period) - tax_emission_regional(model_region, emission, period)) $ (tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period) AND (tax_emission_regional_stoch(stochastic_scenario, model_region, emission, period) < tax_emission_regional(model_region, emission, period)))
             + (tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period) - tax_emission_sectoral(model_region, sector, emission, period)) $ (tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period) AND (tax_emission_sectoral_stoch(stochastic_scenario, model_region, sector, emission, period) < tax_emission_sectoral(model_region, sector, emission, period)))
          )
          * SECTORAL_EMISSION(period, model_region, sector, emission)
          * discount_period(model_region, sector)
         )
* carbon price on cumulative emission for last period
    + SUM(emission,
          (
                (tax_emission_cumulative_stoch(stochastic_scenario, emission) - tax_emission_cumulative(emission)) $ (tax_emission_cumulative_stoch(stochastic_scenario, emission) AND (tax_emission_cumulative_stoch(stochastic_scenario, emission) < tax_emission_cumulative(emission)))
                * CUMULATIVE_EMISSION(emission)
          ) $ last_period(period)
         )
      )
      =E= 0
;

$GOTO total_risk_function_lin_quad

* ------------------------------------------------------------------------------
* Linear Model (original MESSAGE formulation)
* ------------------------------------------------------------------------------

$LABEL total_risk_function_linear

EQ_RISK_FUNCTION_TOTAL
..
      RISK_FUNCTION_TOTAL
* stochastic risk function component
    - 1/CARD(stochastic_scenario)
    * SUM((period, stochastic_scenario) $ (NOT first_period(period)),
          rho * RISK_FUNCTION(period, stochastic_scenario)
          * discount_factor(period, 'world', 'primary_energy')
      )
      =E= 0
;

EQ_OPPORTUNITY_FUNCTION_TOTAL
..
      OPPORTUNITY_FUNCTION_TOTAL
* stochastic risk function component
    - 1/CARD(stochastic_scenario)
    * SUM((period, stochastic_scenario) $ (NOT first_period(period)),
          eta * OPPORTUNITY_FUNCTION(period, stochastic_scenario)
          * discount_factor(period, 'world', 'primary_energy')
      )
      =E= 0
;

$IFI "%type%" == "stochastic_risk_constraint" $GOTO risk_constraint_equation
$IFI "%type%" == "stochastic_cvar_constraint" $GOTO cvar_equation
$IFI "%type%" == "stochastic_cost_constraint_cvar" $GOTO cvar_equation
$IFI "%type%" == "stochastic_linear_comb_cvar" $GOTO cvar_equation
$GOTO risk_function_end

* ------------------------------------------------------------------------------
* Quadratic Model (semi-variance, similar to Markovitz approach)
* ------------------------------------------------------------------------------

$LABEL total_risk_function_quadratic

EQ_RISK_FUNCTION_TOTAL
..
      RISK_FUNCTION_TOTAL
* stochastic risk function component
    - 1/CARD(stochastic_scenario)
    * SUM((period, stochastic_scenario) $ (NOT first_period(period)),
          rho * RISK_FUNCTION(period, stochastic_scenario) * RISK_FUNCTION(period, stochastic_scenario)
          * discount_factor(period, 'world', 'primary_energy')
      )
      =E= 0
;

EQ_OPPORTUNITY_FUNCTION_TOTAL
..
      OPPORTUNITY_FUNCTION_TOTAL
* stochastic risk function component
    - 1/CARD(stochastic_scenario)
    * SUM((period, stochastic_scenario) $ (NOT first_period(period)),
          eta * OPPORTUNITY_FUNCTION(period, stochastic_scenario)
          * discount_factor(period, 'world', 'primary_energy')
      )
      =E= 0

$GOTO risk_function_end

* ------------------------------------------------------------------------------
* Linear-Quadratic Model (Gruebler, Gritsvskii 1999)
* ------------------------------------------------------------------------------

$LABEL total_risk_function_lin_quad

EQ_RISK_FUNCTION_TOTAL
..
      RISK_FUNCTION_TOTAL
* stochastic risk function component
    - 1/CARD(stochastic_scenario)
    * SUM((period, stochastic_scenario) $ (NOT first_period(period)),
          (
            rho * RISK_FUNCTION(period, stochastic_scenario) * RISK_FUNCTION(period, stochastic_scenario)
          + eta * OPPORTUNITY_FUNCTION(period, stochastic_scenario)
          )
          * discount_factor(period, 'world', 'primary_energy')
      )
      =E= 0
;

EQ_OPPORTUNITY_FUNCTION_TOTAL
..
      OPPORTUNITY_FUNCTION_TOTAL
* stochastic risk function component
    - 1/CARD(stochastic_scenario)
    * SUM((period, stochastic_scenario) $ (NOT first_period(period)),
          eta * OPPORTUNITY_FUNCTION(period, stochastic_scenario)
          * discount_factor(period, 'world', 'primary_energy')
      )
      =E= 0
;

$IFI "%type%" == "stochastic_risk_constraint" $GOTO risk_constraint_equation
$IFI "%type%" == "stochastic_cvar_constraint" $GOTO cvar_equation
$IFI "%type%" == "stochastic_cost_constraint_cvar" $GOTO cvar_equation
$IFI "%type%" == "stochastic_linear_comb_cvar" $GOTO cvar_equation
$GOTO risk_function_end

* ------------------------------------------------------------------------------
* Risk Constrained Model
* ------------------------------------------------------------------------------

$LABEL risk_constraint_equation

EQUATION
    EQ_RISK_CONSTRAINT(period)           'risk constraint per period'
    EQ_RISK_CONSTRAINT_TOTAL             'risk constraint on total (discounted) risk'
;

EQ_RISK_CONSTRAINT(period) $
(NOT first_period(period))
..
      1/CARD(stochastic_scenario)
    * SUM(stochastic_scenario, RISK_FUNCTION(period, stochastic_scenario))
    - risk_max(period)
      =L= 0
;

EQ_RISK_CONSTRAINT_TOTAL
..
      RISK_FUNCTION_TOTAL
    - risk_max_total
      =L= 0
;

$GOTO risk_function_end

* ------------------------------------------------------------------------------
* CVaR Model
* ------------------------------------------------------------------------------

$LABEL cvar_equation

* additional variables
VARIABLE
    CVAR_TOTAL                                                     'total CVaR (discounted)'
    CVAR_PERIOD(period)                                            'CVaR for single period'
    ALPHA(period)                                                  'VaR proxy'
    Z(period, stochastic_scenario_all)                             'auxiliary CVaR variable (risk beyond VaR level alpha)'
;

POSITIVE VARIABLE
    Z                                                            'CVaR - VaR by definition positive'
;

* additional equations
EQUATION
    EQ_CVAR_MAX(period)
    EQ_CVAR_MAX_TOTAL
    EQ_CVAR_AUXILIARY(period, stochastic_scenario_all)
;

EQ_CVAR_MAX(period) $
(NOT first_period(period))
..
      ALPHA(period)
    + 1/(1 - beta) * 1/CARD(stochastic_scenario) * SUM(stochastic_scenario, Z(period, stochastic_scenario))
    - CVAR_PERIOD(period)
     =E= 0
;

EQ_CVAR_MAX_TOTAL
..
      SUM(period $ (NOT first_period(period)),
          (ALPHA(period)
        + 1/(1 - beta) * 1/CARD(stochastic_scenario) * SUM(stochastic_scenario, Z(period, stochastic_scenario)))
        * discount_factor(period, 'world', 'primary_energy')
      )
    - CVAR_TOTAL
     =E= 0
;

EQ_CVAR_AUXILIARY(period, stochastic_scenario) $
(NOT first_period(period))
..
* loss function (negative of opportunity function)
      RISK_FUNCTION(period, stochastic_scenario)
* VaR proxy
    - ALPHA(period)
* auxiliary
    - Z(period, stochastic_scenario)
     =L= 0
;

$IFI "%type%" == "stochastic_cvar_constraint" $GOTO cvar_constraint
$GOTO risk_function_end

$LABEL cvar_constraint

EQUATION
    EQ_CVAR_CONSTRAINT_TOTAL
    EQ_CVAR_CONSTRAINT_PERIOD(period)
;


EQ_CVAR_CONSTRAINT_TOTAL
..
    CVAR_TOTAL
  - cvar_max_total
    =L= 0
;

EQ_CVAR_CONSTRAINT_PERIOD(period)
(NOT first_period(period))
..
    CVAR_PERIOD(period)
  - cvar_max(period)
    =L= 0
;

$GOTO risk_function_end

$LABEL risk_function_end

$IFI "%type%" == "stochastic_linear_combination" $GOTO stochastic_linear_combination
$IFI "%type%" == "stochastic_cost_constraint" $GOTO stochastic_cost_constraint
$IFI "%type%" == "stochastic_risk_constraint" $GOTO stochastic_risk_constraint
$IFI "%type%" == "stochastic_cvar_constraint" $GOTO stochastic_risk_constraint
$IFI "%type%" == "stochastic_cost_constraint_cvar" $GOTO stochastic_cost_constraint
$IFI "%type%" == "stochastic_linear_comb_cvar" $GOTO stochastic_linear_comb_cvar

* ------------------------------------------------------------------------------
* Model Type - Stochastic MESSAGE Model Formulation (linear combination)
* ------------------------------------------------------------------------------
$LABEL stochastic_linear_combination

VARIABLE
    OBJ_LINEAR_COMBINATION               'linear combination of stochastic perfect-foresight costs and risk function'
;

EQUATION
    EQ_OBJ_LINEAR_COMBINATION            'objective function - linear combination of stochastic perfect-foresight costs and risk function'
;

EQ_OBJ_LINEAR_COMBINATION
..
      OBJ_LINEAR_COMBINATION
* deterministic cost components
*    - TOTAL_COST_PERFECT_FORESIGHT
    - TOTAL_COST_MESSAGE
* stochastic risk function component
    - rho * RISK_FUNCTION_TOTAL
      =E= 0
;

* stochastic perfect-foresight model definition

MODEL STOCHASTIC_LINEAR_COMBINATION /ALL/ ;

$GOTO end_equation

* ------------------------------------------------------------------------------
* Model Type - linear combination of total system costs and CVaR
* ------------------------------------------------------------------------------
$LABEL stochastic_linear_comb_cvar

VARIABLE
    OBJ_LINEAR_COMBINATION               'linear combination of stochastic perfect-foresight costs and risk function'
;

EQUATION
    EQ_OBJ_LINEAR_COMBINATION            'objective function - linear combination of stochastic perfect-foresight costs and risk function'
;

EQ_OBJ_LINEAR_COMBINATION
..
      OBJ_LINEAR_COMBINATION
* deterministic cost components
*    - TOTAL_COST_PERFECT_FORESIGHT
    - TOTAL_COST_MESSAGE
* stochastic risk function component
    - rho * CVAR_TOTAL
      =E= 0
;

* stochastic perfect-foresight model definition

MODEL STOCHASTIC_LINEAR_COMB_CVAR /ALL/ ;

$GOTO end_equation

* ------------------------------------------------------------------------------
* Model Type - Cost Constrained Risk Minimization
* ------------------------------------------------------------------------------
$LABEL stochastic_cost_constraint

* stochastic perfect-foresight model definition

MODEL STOCHASTIC_COST_CONSTRAINT /ALL/ ;

$GOTO end_equation

* ------------------------------------------------------------------------------
* Model Type - Risk Constrained Cost Minimization (e.g. Stochastic CVaR Model)
* ------------------------------------------------------------------------------
$LABEL stochastic_risk_constraint

* stochastic perfect-foresight model definition

MODEL STOCHASTIC_RISK_CONSTRAINT /ALL/ ;

$GOTO end_equation

* ------------------------------------------------------------------------------
* End of Equations
* ------------------------------------------------------------------------------

$LABEL end_equation

* ------------------------------------------------------------------------------
* RUN perfect-foresight optimization model
* ------------------------------------------------------------------------------

*first_period(period) $ (sameas(period, '2000')) = yes ;

* fix initial resource volumes

STOCK.FX(first_period, model_region, energycarrier, level) $
    (energycarrier_level(energycarrier, level) AND volume(model_region, energycarrier, level))
    = volume(model_region, energycarrier, level) ;

* ------------------------------------------------------------------------------
* code efficiency: technology availability and stochastic set
* ------------------------------------------------------------------------------

* Perfect Foresight availability of regional technologies
technology_regional(period, model_region, technology, year) $
* technology is commercialy available (year >= start)
    (SUM(year_alias $ (order(year_alias) <= order(year)),
          technology_regional_start(model_region, technology, year_alias))
     AND
     (order(period) ge order(year))
     AND
* technologies from past years available if a capacity bound exists, i.e. they have been built in the past
     (((r_bound_capacity(period, model_region, technology, year, 'LO') > 0) OR
       (r_bound_capacity(period, model_region, technology, year, 'FX') > 0) OR
       (r_bound_capacity(period, model_region, technology, year, 'UP') > 0))
      OR
* or from the past if the technological lifetime is not exceeded (adapted for fractional lifetime)
      ((technical_lifetime(model_region, technology, year)
        > (order(period) - order(year)) * period_length)
       AND
       (period_alias(year)
        OR
        ((SUM(period_alias $ (order(period_alias) le order(period)), (r_bound_capacity(period_alias, model_region, technology, year, 'LO') > 0)) OR
          SUM(period_alias $ (order(period_alias) le order(period)), (r_bound_capacity(period_alias, model_region, technology, year, 'UP') > 0)) OR
          SUM(period_alias $ (order(period_alias) le order(period)), (r_bound_capacity(period_alias, model_region, technology, year, 'FX') > 0)))
         AND
         SUM(first_period, (order(year) < order(first_period)))
        )
       )
      )
      OR
* or in the present period
      (order(period) = order(year)))) = yes
;

DISPLAY technology_regional ;

* set stochastic set accoding to sample size
$IF set stochastic stochastic_scenario(stochastic_scenario_all) = YES $ (ORD(stochastic_scenario_all) <= stochastic_realization) ;

* ------------------------------------------------------------------------------
* set external bounds
* ------------------------------------------------------------------------------

* regional capacity lower bound
CAP.LO(period, model_region, technology, year) $
(technology_regional(period, model_region, technology, year) AND
 r_bound_capacity(period, model_region, technology, year, 'LO') AND
 past_presence(period, year))
 = r_bound_capacity(period, model_region, technology, year, 'LO')
;
* regional capacity upper bound
CAP.UP(period, model_region, technology, year) $
(technology_regional(period, model_region, technology, year) AND
 r_bound_capacity(period, model_region, technology, year, 'UP') AND
 past_presence(period, year))
 = r_bound_capacity(period, model_region, technology, year, 'UP')
;
* regional capacity fixed bound
CAP.FX(period, model_region, technology, year) $
(technology_regional(period, model_region, technology, year) AND
 r_bound_capacity(period, model_region, technology, year, 'FX') AND
 past_presence(period, year))
 = r_bound_capacity(period, model_region, technology, year, 'FX')
;

* regional cumulated capacity lower bound
CAP_TOTAL.LO(period, model_region, technology) $
r_bound_capacity_cumulated(period, model_region, technology, 'LO')
 = r_bound_capacity_cumulated(period, model_region, technology, 'LO')
;
* regional cumulated capacity upper bound
CAP_TOTAL.UP(period, model_region, technology) $
r_bound_capacity_cumulated(period, model_region, technology, 'UP')
 = r_bound_capacity_cumulated(period, model_region, technology, 'UP')
;
* regional cumulated capacity fixed bound
CAP_TOTAL.FX(period, model_region, technology) $
r_bound_capacity_cumulated(period, model_region, technology, 'FX')
 = r_bound_capacity_cumulated(period, model_region, technology, 'FX')
;

* regional activity lower bound
ACT.LO(period, model_region, technology, year) $
(technology_regional(period, model_region, technology, year) AND
 r_bound_activity(period, model_region, technology, year, 'LO') AND
 past_presence(period, year))
 = r_bound_activity(period, model_region, technology, year, 'LO')
;
* regional activity upper bound
ACT.UP(period, model_region, technology, year) $
(technology_regional(period, model_region, technology, year) AND
 r_bound_activity(period, model_region, technology, year, 'UP') AND
 past_presence(period, year))
 = r_bound_activity(period, model_region, technology, year, 'UP')
;
* regional activity fixed bound
ACT.FX(period, model_region, technology, year) $
(technology_regional(period, model_region, technology, year) AND
 r_bound_activity(period, model_region, technology, year, 'FX') AND
 past_presence(period, year))
 = r_bound_activity(period, model_region, technology, year, 'FX')
;
* cumulated regional activity lower bound
ACT_TOTAL.LO(period, model_region, technology) $
r_bound_activity_cumulated(period, model_region, technology, 'LO')
 = r_bound_activity_cumulated(period, model_region, technology, 'LO')
;
* cumulated regional activity upper bound
ACT_TOTAL.UP(period, model_region, technology) $
r_bound_activity_cumulated(period, model_region, technology, 'UP')
 = r_bound_activity_cumulated(period, model_region, technology, 'UP')
;
* cumulated regional activity fixed bound
ACT_TOTAL.FX(period, model_region, technology) $
r_bound_activity_cumulated(period, model_region, technology, 'FX')
 = r_bound_activity_cumulated(period, model_region, technology, 'FX')
;
* ------------------------------ emission bounds -------------------------------
* total emission bounds
TOTAL_EMISSION.LO(period, emission) $
bound_emission_total(period, emission, 'LO')
 = bound_emission_total(period, emission, 'LO')
;
TOTAL_EMISSION.UP(period, emission) $
bound_emission_total(period, emission, 'UP')
 = bound_emission_total(period, emission, 'UP')
;
TOTAL_EMISSION.FX(period, emission) $
bound_emission_total(period, emission, 'FX')
 = bound_emission_total(period, emission, 'FX')
;
* regional emission bounds
REGIONAL_EMISSION.LO(period, model_region, emission) $
bound_emission_regional(period, model_region, emission, 'LO')
 = bound_emission_regional(period, model_region, emission, 'LO')
;
REGIONAL_EMISSION.UP(period, model_region, emission) $
bound_emission_regional(period, model_region, emission, 'UP')
 = bound_emission_regional(period, model_region, emission, 'UP')
;
REGIONAL_EMISSION.FX(period, model_region, emission) $
bound_emission_regional(period, model_region, emission, 'FX')
 = bound_emission_regional(period, model_region, emission, 'FX')
;
* sectoral emission bounds
SECTORAL_EMISSION.LO(period, model_region, sector, emission) $
bound_emission_sectoral(period, model_region, sector, emission, 'LO')
 = bound_emission_sectoral(period, model_region, sector, emission, 'LO')
;
SECTORAL_EMISSION.UP(period, model_region, sector, emission) $
bound_emission_sectoral(period, model_region, sector, emission, 'UP')
 = bound_emission_sectoral(period, model_region, sector, emission, 'UP')
;
SECTORAL_EMISSION.FX(period, model_region, sector, emission) $
bound_emission_sectoral(period, model_region, sector, emission, 'FX')
 = bound_emission_sectoral(period, model_region, sector, emission, 'FX')
;
* ------------------------- energy production bounds ---------------------------
* total energy production bounds
TOTAL_ENERGY_PRODUCTION.LO(period, energycarrier) $
bound_production_total(period, energycarrier, 'LO')
 = bound_production_total(period, energycarrier, 'LO')
;
TOTAL_ENERGY_PRODUCTION.UP(period, energycarrier) $
bound_production_total(period, energycarrier, 'UP')
 = bound_production_total(period, energycarrier, 'UP')
;
TOTAL_ENERGY_PRODUCTION.FX(period, energycarrier) $
bound_production_total(period, energycarrier, 'FX')
 = bound_production_total(period, energycarrier, 'FX')
;
* regional energy production bounds
REGIONAL_ENERGY_PRODUCTION.LO(period, model_region, energycarrier) $
bound_production_regional(period, model_region, energycarrier, 'LO')
 = bound_production_regional(period, model_region, energycarrier, 'LO')
;
REGIONAL_ENERGY_PRODUCTION.UP(period, model_region, energycarrier) $
bound_production_regional(period, model_region, energycarrier, 'UP')
 = bound_production_regional(period, model_region, energycarrier, 'UP')
;
REGIONAL_ENERGY_PRODUCTION.FX(period, model_region, energycarrier) $
bound_production_regional(period, model_region, energycarrier, 'FX')
 = bound_production_regional(period, model_region, energycarrier, 'FX')
;
* sectoral energy production bounds
SECTORAL_ENERGY_PRODUCTION.LO(period, model_region, sector, energycarrier) $
bound_production_sectoral(period, model_region, sector, energycarrier, 'LO')
 = bound_production_sectoral(period, model_region, sector, energycarrier, 'LO')
;
SECTORAL_ENERGY_PRODUCTION.UP(period, model_region, sector, energycarrier) $
bound_production_sectoral(period, model_region, sector, energycarrier, 'UP')
 = bound_production_sectoral(period, model_region, sector, energycarrier, 'UP')
;
SECTORAL_ENERGY_PRODUCTION.FX(period, model_region, sector, energycarrier) $
bound_production_sectoral(period, model_region, sector, energycarrier, 'FX')
 = bound_production_sectoral(period, model_region, sector, energycarrier, 'FX')
;
* -------------------------- energy consumption bounds -------------------------
* total energy consumption bounds
TOTAL_ENERGY_CONSUMPTION.LO(period, energycarrier) $
bound_consumption_total(period, energycarrier, 'LO')
 = bound_consumption_total(period, energycarrier, 'LO')
;
TOTAL_ENERGY_CONSUMPTION.UP(period, energycarrier) $
bound_consumption_total(period, energycarrier, 'UP')
 = bound_consumption_total(period, energycarrier, 'UP')
;
TOTAL_ENERGY_CONSUMPTION.FX(period, energycarrier) $
bound_consumption_total(period, energycarrier, 'FX')
 = bound_consumption_total(period, energycarrier, 'FX')
;
* regional energy consumption bounds
REGIONAL_ENERGY_CONSUMPTION.LO(period, model_region, energycarrier) $
bound_consumption_regional(period, model_region, energycarrier, 'LO')
 = bound_consumption_regional(period, model_region, energycarrier, 'LO')
;
REGIONAL_ENERGY_CONSUMPTION.UP(period, model_region, energycarrier) $
bound_consumption_regional(period, model_region, energycarrier, 'UP')
 = bound_consumption_regional(period, model_region, energycarrier, 'UP')
;
REGIONAL_ENERGY_CONSUMPTION.FX(period, model_region, energycarrier) $
bound_consumption_regional(period, model_region, energycarrier, 'FX')
 = bound_consumption_regional(period, model_region, energycarrier, 'FX')
;
* sectoral energy consumption bounds
SECTORAL_ENERGY_CONSUMPTION.LO(period, model_region, sector, energycarrier) $
bound_consumption_sectoral(period, model_region, sector, energycarrier, 'LO')
 = bound_consumption_sectoral(period, model_region, sector, energycarrier, 'LO')
;
SECTORAL_ENERGY_CONSUMPTION.UP(period, model_region, sector, energycarrier) $
bound_consumption_sectoral(period, model_region, sector, energycarrier, 'UP')
 = bound_consumption_sectoral(period, model_region, sector, energycarrier, 'UP')
;
SECTORAL_ENERGY_CONSUMPTION.FX(period, model_region, sector, energycarrier) $
bound_consumption_sectoral(period, model_region, sector, energycarrier, 'FX')
 = bound_consumption_sectoral(period, model_region, sector, energycarrier, 'FX')
;

* ------------------------------------------------------------------------------
* solve statements
* ------------------------------------------------------------------------------

$IFI "%type%" == "deterministic" $GOTO solve_deterministic
$IFI "%type%" == "stochastic_linear_combination" $GOTO solve_linear_combination
$IFI "%type%" == "stochastic_cost_constraint" $GOTO solve_deterministic
$IFI "%type%" == "stochastic_risk_constraint" $GOTO solve_risk_constraint
$IFI "%type%" == "stochastic_cvar_constraint" $GOTO solve_risk_constraint
$IFI "%type%" == "stochastic_cost_constraint_cvar" $GOTO solve_deterministic
$IFI "%type%" == "stochastic_linear_comb_cvar" $GOTO solve_linear_comb_cvar

* ------------------------------------------------------------------------------
* solve deterministic perfect-foresight model
* ------------------------------------------------------------------------------

$LABEL solve_deterministic

DETERMINISTIC.optfile = 1 ;
DETERMINISTIC.holdfixed = 1 ;

SOLVE DETERMINISTIC using LP minimize TOTAL_COST_MESSAGE ;

cost_base_total = TOTAL_COST_MESSAGE.L ;
cost_base(period) = TOTAL_COST_PERIOD_MESSAGE.L(period) ;

$IFI "%type%" == "stochastic_cost_constraint" $GOTO solve_cost_constraint
$IFI "%type%" == "stochastic_cost_constraint_cvar" $GOTO solve_cost_constraint_cvar
$GOTO solve_end

* ------------------------------------------------------------------------------
* solve stochastic perfect-foresight model with linear combination of total system costs and risk-function
* ------------------------------------------------------------------------------

$LABEL solve_linear_combination

STOCHASTIC_LINEAR_COMBINATION.optfile = 1 ;
STOCHASTIC_LINEAR_COMBINATION.holdfixed = 1 ;

SOLVE STOCHASTIC_LINEAR_COMBINATION using LP minimize OBJ_LINEAR_COMBINATION ;

$GOTO solve_end

* ------------------------------------------------------------------------------
* solve stochastic perfect-foresight model with linear combination of total system costs and beta-CVaR
* ------------------------------------------------------------------------------

$LABEL solve_linear_comb_cvar

STOCHASTIC_LINEAR_COMB_CVAR.optfile = 1 ;
STOCHASTIC_LINEAR_COMB_CVAR.holdfixed = 1 ;

SOLVE STOCHASTIC_LINEAR_COMB_CVAR using LP minimize OBJ_LINEAR_COMBINATION ;

$GOTO solve_end

* ------------------------------------------------------------------------------
* solve cost-constrained stochastic perfect-foresight model
* ------------------------------------------------------------------------------

$LABEL solve_cost_constraint
DISPLAY f ;
* set upper bound on expected value for costs (risk premium)
*TOTAL_COST_MESSAGE.UP = (1 + f) * cost_base_total ;
* set period-specific upper bound on expected value for costs (risk premium)
f_period(period) = f ;
TOTAL_COST_PERIOD_MESSAGE.UP(period) $ (NOT first_period(period)) = (1 + f_period(period)) * cost_base(period) ;

STOCHASTIC_COST_CONSTRAINT.optfile = 1 ;
STOCHASTIC_COST_CONSTRAINT.holdfixed = 1 ;

$IFI "%formulation%" == "linear" $GOTO solve_cost_linear
$IFI "%formulation%" == "quadratic" $GOTO solve_cost_quadratic
$IFI "%formulation%" == "linear_quadratic" $GOTO solve_cost_linear_quadratic

* linear model

$LABEL solve_cost_linear

SOLVE STOCHASTIC_COST_CONSTRAINT using LP minimize RISK_FUNCTION_TOTAL ;

$GOTO solve_end

* quadratic or linear-quadratic model

$LABEL solve_cost_quadratic
$LABEL solve_cost_linear_quadratic

SOLVE STOCHASTIC_COST_CONSTRAINT using QCP minimize RISK_FUNCTION_TOTAL ;

$GOTO solve_end

* ------------------------------------------------------------------------------
* solve cost-constrained stochastic CVaR perfect-foresight model
* ------------------------------------------------------------------------------

$LABEL solve_cost_constraint_cvar
* set upper bound on expected value for costs (risk premium)
*TOTAL_COST_MESSAGE.UP = (1 + f) * cost_base_total ;
* set period-specific upper bound on expected value for costs (risk premium)
f_period(period) = f ;
TOTAL_COST_PERIOD_MESSAGE.UP(period) $ (NOT first_period(period)) = (1 + f_period(period)) * cost_base(period) ;

STOCHASTIC_COST_CONSTRAINT.optfile = 1 ;
STOCHASTIC_COST_CONSTRAINT.holdfixed = 1 ;

* only defined for linear model (possibly to be extended)

SOLVE STOCHASTIC_COST_CONSTRAINT using LP minimize CVAR_TOTAL ;

$GOTO solve_end

* ------------------------------------------------------------------------------
* solve risk-constrained stochastic perfect-foresight model
* ------------------------------------------------------------------------------

$LABEL solve_risk_constraint

STOCHASTIC_RISK_CONSTRAINT.optfile = 1 ;
STOCHASTIC_RISK_CONSTRAINT.holdfixed = 1 ;

$IFI "%formulation%" == "linear" $GOTO solve_risk_linear
$IFI "%formulation%" == "quadratic" $GOTO solve_risk_quadratic
$IFI "%formulation%" == "linear_quadratic" $GOTO solve_risk_linear_quadratic

* linear model

$LABEL solve_risk_linear

SOLVE STOCHASTIC_RISK_CONSTRAINT using LP minimize TOTAL_COST_MESSAGE ;

$GOTO solve_end

* quadratic or linear-quadratic model

$LABEL solve_risk_quadratic
$LABEL solve_risk_linear_quadratic

SOLVE STOCHASTIC_RISK_CONSTRAINT using QCP minimize TOTAL_COST_MESSAGE ;

$GOTO solve_end

* ------------------------------------------------------------------------------
* end solve
* ------------------------------------------------------------------------------

$LABEL solve_end


* ------------------------------------------------------------------------------
* report writing of results
* ------------------------------------------------------------------------------

$IF NOT set id $SET filename "%title%_%type%_%formulation%_%stochastic%"
$IF set id $SET filename "%title%_%id%_%type%_%formulation%_%stochastic%"
$IF NOT set stochastic $SET filename "%title%_%type%"

$IF NOT set stochastic $GOTO report_writing
* ------------------------------------------------------------------------------
* sort risk measure distribution to extract quantiles, etc.
* GAMS sorting algorithm (adopted from original code by Thomas F. Rutherford)
* ------------------------------------------------------------------------------

PARAMETER risk_value(stochastic_scenario_all) 'risk function values to be sorted' ;

* calculate risk-function values
$IF "%formulation%" == "linear" risk_value(stochastic_scenario) = SUM(period $ (NOT first_period(period)), RISK_FUNCTION.L(period, stochastic_scenario) * discount_factor(period, 'world', 'primary_energy')) ;
$IF "%formulation%" == "quadratic" risk_value(stochastic_scenario) = SUM(period $ (NOT first_period(period)), RISK_FUNCTION.L(period, stochastic_scenario) * RISK_FUNCTION.L(period, stochastic_scenario) * discount_factor(period, 'world', 'primary_energy')) ;
$IF "%formulation%" == "linear_quadratic" risk_value(stochastic_scenario) = SUM(period $ (NOT first_period(period)), (rho * RISK_FUNCTION.L(period, stochastic_scenario) * RISK_FUNCTION.L(period, stochastic_scenario) + eta * OPPORTUNITY_FUNCTION.L(period, stochastic_scenario)) * discount_factor(period, 'world', 'primary_energy')) ;

ALIAS (stochastic_scenario, stochastic_scenario_alias) ;

SET stochastic_scenario_ordered(stochastic_scenario_all) 'ordered set of stochastic scenarios' ;
stochastic_scenario_ordered(stochastic_scenario_all) = YES $ (ORD(stochastic_scenario_all) <= stochastic_realization) ;

PARAMETER RANK(stochastic_scenario_all) 'rank order of risk function values' ;

* sorting algorithm
RANK(stochastic_scenario) = SUM(stochastic_scenario_alias$(risk_value(stochastic_scenario_alias) GE risk_value(stochastic_scenario)), 1) ;

* renormalize expected value to exclude zeros
SCALAR n_gt_0    'renormalization factor' ;

n_gt_0 = SUM(stochastic_scenario $ (risk_value(stochastic_scenario) > 0), 1) ;

DISPLAY n_gt_0 ;

* ------------------------------------------------------------------------------
* report writing (MESSAGE-compatible output tables)
* ------------------------------------------------------------------------------
$LABEL report_writing

FILE model_output / "%filename%.csv" / ;

* options for output format
model_output.lw = 0 ;
model_output.sw = 0 ;
* numeric field width
model_output.nw = 12 ;
* scientific notation
model_output.nr = 2 ;
* number of decimals
model_output.nd = 4 ;
* page control
model_output.pc = 2 ;

PUT model_output ;

* File Header contains scenario title and execution date and time
model_output.pc = 2 ;
PUT system.title/ system.date, ' ', system.time/ ;

* data output mode for csv-type file
model_output.pc = 5 ;

PUT model_output ;

* ------------------------------------------------------------------------------
* electricity generation
* ------------------------------------------------------------------------------

PUT / 'electricity generation', 'TWyr'/ ;

LOOP(model_region,
        PUT '', 'Coal', 'Gas', 'Nuclear', 'Biomass', 'Hydro', 'Wind', 'Solar', 'CoalCCS', 'GasCCS', 'BiomassCCS' / ;
    LOOP(period,
        PUT period.tl,
            (ACT_TOTAL.L(period, model_region, 'coal_ppl') - ACT_TOTAL.L(period, model_region, 'c_ppl_co2scr')),
            (ACT_TOTAL.L(period, model_region, 'gas_cc') - ACT_TOTAL.L(period, model_region, 'g_ppl_co2scr')),
             ACT_TOTAL.L(period, model_region, 'nuc_hc'),
            (ACT_TOTAL.L(period, model_region, 'bio_ppl') - ACT_TOTAL.L(period, model_region, 'bio_ppl_co2scr')),
             ACT_TOTAL.L(period, model_region, 'hydro_hc'),
            (ACT_TOTAL.L(period, model_region, 'wind_ppl') + ACT_TOTAL.L(period, model_region, 'wind_ppl2')),
             ACT_TOTAL.L(period, model_region, 'solar_pv_ppl'),
            ((1 - input(model_region, 'c_ppl_co2scr', 'electricity', 'secondary', period)) * ACT_TOTAL.L(period, model_region, 'c_ppl_co2scr')),
            ((1 - input(model_region, 'g_ppl_co2scr', 'electricity', 'secondary', period)) * ACT_TOTAL.L(period, model_region, 'g_ppl_co2scr')),
            ((1 - input(model_region, 'bio_ppl_co2scr', 'electricity', 'secondary', period)) * ACT_TOTAL.L(period, model_region, 'bio_ppl_co2scr')) /
    ) ;
) ;

* ------------------------------------------------------------------------------
* other conversion technologies
* ------------------------------------------------------------------------------

PUT / 'other conversion', 'TWyr'/ ;

LOOP(model_region,
        PUT '', 'h2', 'ctl', 'gtl', 'btl', 'refinery' / ;
    LOOP(period,
        PUT period.tl,
            ACT_TOTAL.L(period, model_region, 'h2_elec'),
            ACT_TOTAL.L(period, model_region, 'meth_coal'),
            ACT_TOTAL.L(period, model_region, 'meth_ng'),
            ACT_TOTAL.L(period, model_region, 'eth_bio'),
            ACT_TOTAL.L(period, model_region, 'ref_lol') /
    ) ;
) ;

* ------------------------------------------------------------------------------
* total primary energy supply - extended
* ------------------------------------------------------------------------------

PUT / 'total primary energy supply (TPES) - extended', 'TWyr'/ ;

LOOP(model_region,
        PUT '', 'coal', 'oil.conv', 'oil.unconv', 'gas.conv', 'gas.unconv', 'nuclear', 'biomass', 'hydro', 'wind', 'solar' / ;
    LOOP(period,
        PUT period.tl,
            ENERGY_SUPPLY.L(period, model_region, 'coal', 'primary'),
*            ENERGY_SUPPLY.L(period, model_region, 'oil', 'primary'),
*            ENERGY_SUPPLY.L(period, model_region, 'gas', 'primary'),
            (ACT_TOTAL.L(period, model_region, 'oil_extr_1') * output(model_region, 'oil_extr_1', 'oil', 'primary', period)),
            (ACT_TOTAL.L(period, model_region, 'oil_extr_4') * output(model_region, 'oil_extr_4', 'oil', 'primary', period)),
            (ACT_TOTAL.L(period, model_region, 'gas_extr_1') * output(model_region, 'gas_extr_1', 'gas', 'primary', period)),
            (ACT_TOTAL.L(period, model_region, 'gas_extr_5') * output(model_region, 'gas_extr_5', 'gas', 'primary', period)),
            ENERGY_SUPPLY.L(period, model_region, 'nuclear', 'primary'),
            ENERGY_SUPPLY.L(period, model_region, 'biomass', 'primary'),
            ENERGY_SUPPLY.L(period, model_region, 'hydro', 'primary'),
            ENERGY_SUPPLY.L(period, model_region, 'wind', 'primary'),
            ENERGY_SUPPLY.L(period, model_region, 'solar', 'primary') /
    ) ;
) ;

* ------------------------------------------------------------------------------
* final energy consumption
* ------------------------------------------------------------------------------

PUT / 'final energy consumption (FEC)', 'TWyr'/ ;

LOOP(model_region,
        PUT '', 'coal', 'gas', 'liquids', 'biomass', 'electricity' / ;
    LOOP(period,
        PUT period.tl,
            SECTORAL_ENERGY_CONSUMPTION.L(period, model_region, 'end_use', 'coal'),
            SECTORAL_ENERGY_CONSUMPTION.L(period, model_region, 'end_use', 'gas'),
            SECTORAL_ENERGY_CONSUMPTION.L(period, model_region, 'end_use', 'liquids'),
            SECTORAL_ENERGY_CONSUMPTION.L(period, model_region, 'end_use', 'biomass'),
            SECTORAL_ENERGY_CONSUMPTION.L(period, model_region, 'end_use', 'electricity') /
    ) ;
) ;

* ------------------------------------------------------------------------------
* energy-saving
* ------------------------------------------------------------------------------

PUT / 'energy-saving', 'TWyr'/ ;

LOOP(model_region,
        PUT '', 'electric std', 'electric save1', 'electric save2', 'non-electric std', 'non-electric save1', 'non-electric save2', 'transport std', 'transport save1', 'transport save2' / ;
    LOOP(period,
        PUT period.tl,
            (ACT_TOTAL.L(period, model_region, 'enduse_electric_std') * (1 - input(model_region, 'enduse_electric_std', 'electricity', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_electric_save1') * (1 - input(model_region, 'enduse_electric_save1', 'electricity', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_electric_save2') * (1 - input(model_region, 'enduse_electric_save1', 'electricity', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_non_electric_std') * (1 - input(model_region, 'enduse_non_electric_std', 'non_electric', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_non_electric_save1') * (1 - input(model_region, 'enduse_non_electric_save1', 'non_electric', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_non_electric_save2') * (1 - input(model_region, 'enduse_non_electric_save2', 'non_electric', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_transport_std') * (1 - input(model_region, 'enduse_transport_std', 'transport', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_transport_save1') * (1 - input(model_region, 'enduse_transport_save1', 'transport', 'final', period))),
            (ACT_TOTAL.L(period, model_region, 'enduse_transport_save2') * (1 - input(model_region, 'enduse_transport_save2', 'transport', 'final', period))) /
    ) ;
) ;

* ------------------------------------------------------------------------------
* shadow prices
* ------------------------------------------------------------------------------

PUT / 'energy carrier shadow prices', '$/kWyr'/ ;

LOOP(model_region,
        PUT '', 'coal', 'oil', 'gas', 'nuclear', 'biomass', 'hydro', 'wind', 'solar', 'electricity g', 'liquids g', 'electricity f', 'liquids f', 'solids f', 'gas f', 'non-electric f', 'transport f' / ;
    LOOP(period,
        PUT period.tl,
            (EQ_ENERGY_BALANCE.M(period, model_region, 'coal', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'oil', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'gas', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'nuclear', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'biomass', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'hydro', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'wind', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'solar', 'primary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'electricity', 'secondary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'liquids', 'secondary')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'electricity', 'final')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'liquids', 'final')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'solids', 'final')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'gas', 'final')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'non_electric', 'final')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy')),
            (EQ_ENERGY_BALANCE.M(period, model_region, 'transport', 'final')/discount_period(model_region, 'primary_energy')/discount_factor(period, model_region, 'primary_energy'))/
    ) ;
) ;

* ------------------------------------------------------------------------------
* resources
* ------------------------------------------------------------------------------

PUT / 'resources (at end of period)', 'ZJ'/ ;

LOOP(model_region,
        PUT '', 'coal', 'conv. oil', 'unconv. oil', 'conv. gas', 'unconv. gas' / ;
    LOOP(period,
        PUT period.tl,
            ((STOCK.L(period, model_region, 'coal', 'resource') - input(model_region, 'coal_extr', 'coal', 'resource', period) * ACT_TOTAL.L(period, model_region, 'coal_extr')) * 0.03154),
            ((STOCK.L(period, model_region, 'oil_conventional', 'resource') - input(model_region, 'oil_extr_1', 'oil_conventional', 'resource', period) * ACT_TOTAL.L(period, model_region, 'oil_extr_1')) * 0.03154),
            ((STOCK.L(period, model_region, 'oil_unconventional', 'resource') - input(model_region, 'oil_extr_4', 'oil_unconventional', 'resource', period) * ACT_TOTAL.L(period, model_region, 'oil_extr_4')) * 0.03154),
            ((STOCK.L(period, model_region, 'gas_conventional', 'resource') - input(model_region, 'gas_extr_1', 'gas_conventional', 'resource', period) * ACT_TOTAL.L(period, model_region, 'gas_extr_1')) * 0.03154),
            ((STOCK.L(period, model_region, 'gas_unconventional', 'resource') - input(model_region, 'gas_extr_5', 'gas_unconventional', 'resource', period) * ACT_TOTAL.L(period, model_region, 'gas_extr_5')) * 0.03154) /
    ) ;
) ;

* ------------------------------------------------------------------------------
* emissions
* ------------------------------------------------------------------------------

PUT / 'emissions', 'Gt C'/ ;

LOOP(model_region,
        PUT '', 'primary', 'conversion', 'end-use', 'total' / ;
    LOOP(period,
        PUT period.tl,
            SECTORAL_EMISSION.L(period, model_region, 'primary_energy', 'CO2'),
            SECTORAL_EMISSION.L(period, model_region, 'conversion', 'CO2'),
            SECTORAL_EMISSION.L(period, model_region, 'end_use', 'CO2'),
            REGIONAL_EMISSION.L(period, model_region, 'CO2') /
    ) ;
) ;

* ------------------------------------------------------------------------------
* CCS
* ------------------------------------------------------------------------------

PUT / 'CCS: captured emissions', 'Gt C'/ ;

LOOP(model_region,
        PUT '', 'gas ccs', 'coal ccs', 'biomass ccs' / ;
    LOOP(period,
        PUT period.tl,
            SUM(year $ past_presence(period, year), ACT.L(period, model_region, 'g_ppl_co2scr', year) * emission_output(model_region, 'g_ppl_co2scr', 'CO2', year)),
            SUM(year $ past_presence(period, year), ACT.L(period, model_region, 'c_ppl_co2scr', year) * emission_output(model_region, 'c_ppl_co2scr', 'CO2', year)),
            SUM(year $ past_presence(period, year), ACT.L(period, model_region, 'bio_ppl_co2scr', year) * emission_output(model_region, 'bio_ppl_co2scr', 'CO2', year)) /
    ) ;
) ;

* ------------------------------------------------------------------------------
* costs, risks, opportunities, ...
* ------------------------------------------------------------------------------

PUT / 'costs', 'T$'/ ;

PUT '', 'cost', 'E[risk]', 'E[opportunity]', 'risk 50th', 'risk 84.1th' 'risk 90th', 'risk 95th', 'risk 99th', 'VaR', 'CVaR' / ;
LOOP(period,
    PUT period.tl,
        SUM((model_region, sector) $ regional_sector(model_region, sector), SECTORAL_COST_MESSAGE.L(period, model_region, sector))
$IF "%formulation%" == "linear" , (SUM(stochastic_scenario, RISK_FUNCTION.L(period, stochastic_scenario))/CARD(stochastic_scenario))
$IF "%formulation%" == "linear" , 0
$IF "%formulation%" == "quadratic" , (SUM(stochastic_scenario, RISK_FUNCTION.L(period, stochastic_scenario) * RISK_FUNCTION.L(period, stochastic_scenario))/CARD(stochastic_scenario))
$IF "%formulation%" == "quadratic" , 0
$IF "%formulation%" == "linear_quadratic" , (SUM(stochastic_scenario, RISK_FUNCTION.L(period, stochastic_scenario) * RISK_FUNCTION.L(period, stochastic_scenario))/CARD(stochastic_scenario))
$IF "%formulation%" == "linear_quadratic" , (SUM(stochastic_scenario, OPPORTUNITY_FUNCTION.L(period, stochastic_scenario))/CARD(stochastic_scenario))
$IF "%type%" == "stochastic_cvar_constraint" , '', '', '', '', '', ALPHA.L(period)
$IF "%type%" == "stochastic_cvar_constraint" , (ALPHA.L(period) + 1/(1 - beta) * SUM(stochastic_scenario, Z.L(period, stochastic_scenario))/CARD(stochastic_scenario))
$IF "%type%" == "stochastic_cost_constraint_cvar" , '', '', '', '', '', ALPHA.L(period)
$IF "%type%" == "stochastic_cost_constraint_cvar" , (ALPHA.L(period) + 1/(1 - beta) * SUM(stochastic_scenario, Z.L(period, stochastic_scenario))/CARD(stochastic_scenario))
$IF "%type%" == "stochastic_linear_comb_cvar" , '', '', '', '', '', ALPHA.L(period)
$IF "%type%" == "stochastic_linear_comb_cvar" , (ALPHA.L(period) + 1/(1 - beta) * SUM(stochastic_scenario, Z.L(period, stochastic_scenario))/CARD(stochastic_scenario))
        / ;
) ;
PUT "discounted",
TOTAL_COST_MESSAGE.L
* just costs of final period:
$IF "%formulation%" == "linear" , RISK_FUNCTION_TOTAL.L
$IF "%formulation%" == "linear" , 0
$IF "%formulation%" == "quadratic" , RISK_FUNCTION_TOTAL.L
$IF "%formulation%" == "quadratic" , 0
$IF "%formulation%" == "linear_quadratic" , (SUM((stochastic_scenario, period) $ (NOT first_period(period)), RISK_FUNCTION.L(period, stochastic_scenario) * RISK_FUNCTION.L(period, stochastic_scenario) * discount_factor(period, 'world', 'primary_energy'))/CARD(stochastic_scenario))
$IF "%formulation%" == "linear_quadratic" , (SUM((stochastic_scenario, period) $ (NOT first_period(period)), OPPORTUNITY_FUNCTION.L(period, stochastic_scenario) * discount_factor(period, 'world', 'primary_energy'))/CARD(stochastic_scenario))
$IF NOT set stochastic $GOTO skip_percentiles
* percentiles
* 50th percentile does sometimes (mostly in linear case?) not exist -- think of work-around
LOOP(stochastic_scenario $ (RANK(stochastic_scenario) EQ round(0.5*CARD(stochastic_scenario))),
         PUT , risk_value(stochastic_scenario)
) ;
* if no 50th percentile not found set value to zero
IF (SUM(stochastic_scenario, RANK(stochastic_scenario) EQ round(0.5*CARD(stochastic_scenario))) EQ 0,
         PUT , 0
);
LOOP((stochastic_scenario)$(RANK(stochastic_scenario) EQ round(0.159*CARD(stochastic_scenario))),
         PUT , risk_value(stochastic_scenario)
) ;
LOOP((stochastic_scenario)$(RANK(stochastic_scenario) EQ round(0.1*CARD(stochastic_scenario))),
         PUT , risk_value(stochastic_scenario)
) ;
LOOP((stochastic_scenario)$(RANK(stochastic_scenario) EQ round(0.05*CARD(stochastic_scenario))),
         PUT , risk_value(stochastic_scenario)
) ;
LOOP((stochastic_scenario)$(RANK(stochastic_scenario) EQ round(0.01*CARD(stochastic_scenario))),
         PUT , risk_value(stochastic_scenario)
) ;
$LABEL skip_percentiles
$IF "%type%" == "stochastic_cvar_constraint" PUT , (SUM(period $ (NOT first_period(period)), ALPHA.L(period) * discount_factor(period, 'world', 'primary_energy')))
$IF "%type%" == "stochastic_cvar_constraint" PUT , (SUM(period $ (NOT first_period(period)), (ALPHA.L(period) +  1/(1 - beta) * SUM(stochastic_scenario, Z.L(period, stochastic_scenario))/CARD(stochastic_scenario)) * discount_factor(period, 'world', 'primary_energy')))
$IF "%type%" == "stochastic_cost_constraint_cvar" PUT , (SUM(period $ (NOT first_period(period)), ALPHA.L(period) * discount_factor(period, 'world', 'primary_energy')))
$IF "%type%" == "stochastic_cost_constraint_cvar" PUT , (SUM(period $ (NOT first_period(period)), (ALPHA.L(period) +  1/(1 - beta) * SUM(stochastic_scenario, Z.L(period, stochastic_scenario))/CARD(stochastic_scenario)) * discount_factor(period, 'world', 'primary_energy')))
$IF "%type%" == "stochastic_linear_comb_cvar" PUT , (SUM(period $ (NOT first_period(period)), ALPHA.L(period) * discount_factor(period, 'world', 'primary_energy')))
$IF "%type%" == "stochastic_linear_comb_cvar" PUT , (SUM(period $ (NOT first_period(period)), (ALPHA.L(period) +  1/(1 - beta) * SUM(stochastic_scenario, Z.L(period, stochastic_scenario))/CARD(stochastic_scenario)) * discount_factor(period, 'world', 'primary_energy')))
PUT /

PUTCLOSE model_output ;
